import dataclasses
import numbers
from unittest.mock import AsyncMock, patch

import pytest

from starknet_py.constants import EXPECTED_RPC_VERSION, STRK_FEE_CONTRACT_ADDRESS
from starknet_py.hash.selector import get_selector_from_name
from starknet_py.net.client_errors import ClientError
from starknet_py.net.client_models import (
    BlockHeader,
    BlockStatus,
    Call,
    ContractsStorageKeys,
    DAMode,
    DeclareTransactionV3,
    DeployAccountTransactionV3,
    EmittedEvent,
    EstimatedFee,
    EventsChunk,
    InvokeTransactionV3,
    PreConfirmedBlockHeader,
    PreConfirmedStarknetBlockWithReceipts,
    ResourceBounds,
    ResourceBoundsMapping,
    StarknetBlock,
    StarknetBlockWithReceipts,
    Transaction,
    TransactionExecutionStatus,
    TransactionFinalityStatus,
    TransactionReceipt,
    TransactionStatus,
)
from starknet_py.net.executable_models import (
    CasmClass,
    Deref,
    Immediate,
    TestLessThan,
    TestLessThanOrEqual,
)
from starknet_py.net.full_node_client import FullNodeClient
from starknet_py.net.http_client import IncompatibleRPCVersionWarning, RpcHttpClient
from starknet_py.net.models import StarknetChainId
from starknet_py.net.networks import SEPOLIA, default_token_address_for_network
from starknet_py.tests.e2e.fixtures.constants import (
    EMPTY_CONTRACT_ADDRESS_SEPOLIA,
    MAX_RESOURCE_BOUNDS_SEPOLIA,
    STRK_CLASS_HASH,
)
from starknet_py.transaction_errors import TransactionRevertedError


@pytest.mark.parametrize(
    "transaction_hash",
    (
        "0x016df225d14eb927b1c85ec85d2f9f4fc7653ba13a99e30ffe9e21c96ddc7a6d",  # invoke
        "0x0255f63b1dbd52902e2fb5707d2d2b52d5600fa228f0655b02b78bfcf9cab353",  # declare
        "0x0379c52f40fad2d94152d7c924b69cd61a99cf45b85ba9cb836f69026db67af8",  # deploy_account
        "0x06098d74f3fe1b2b96dcfbb3b9ca9be0c396bde0a0825e111fcbefec9c34fcc6",  # l1_handler
    ),
)
@pytest.mark.asyncio
async def test_get_transaction_receipt(client_sepolia_testnet, transaction_hash):
    receipt = await client_sepolia_testnet.get_transaction_receipt(
        tx_hash=transaction_hash
    )

    assert isinstance(receipt, TransactionReceipt)
    assert receipt.execution_status is not None
    assert receipt.finality_status is not None
    assert receipt.execution_resources is not None
    assert receipt.type is not None


@pytest.mark.skip("TODO(#1651)")
@pytest.mark.asyncio
async def test_wait_for_tx_reverted(account_sepolia_testnet):
    account = account_sepolia_testnet
    # Calldata too long for the function (it has no parameters) to trigger REVERTED status
    call = Call(
        to_addr=int(EMPTY_CONTRACT_ADDRESS_SEPOLIA, 0),
        selector=get_selector_from_name("empty"),
        calldata=[0x1, 0x2, 0x3, 0x4, 0x5],
    )
    sign_invoke = await account.sign_invoke_v3(
        calls=call, resource_bounds=MAX_RESOURCE_BOUNDS_SEPOLIA
    )
    invoke = await account.client.send_transaction(sign_invoke)

    with pytest.raises(TransactionRevertedError, match="Input too long for arguments"):
        await account.client.wait_for_tx(tx_hash=invoke.transaction_hash)


@pytest.mark.asyncio
async def test_wait_for_tx_accepted(account_sepolia_testnet):
    account = account_sepolia_testnet
    call = Call(
        to_addr=int(EMPTY_CONTRACT_ADDRESS_SEPOLIA, 0),
        selector=get_selector_from_name("empty"),
        calldata=[],
    )
    sign_invoke = await account.sign_invoke_v3(calls=call, auto_estimate=True)
    invoke = await account.client.send_transaction(sign_invoke)

    result = await account.client.wait_for_tx(tx_hash=invoke.transaction_hash)

    assert result.execution_status == TransactionExecutionStatus.SUCCEEDED
    assert result.finality_status == TransactionFinalityStatus.ACCEPTED_ON_L2


@pytest.mark.asyncio
async def test_sign_invoke_v3_auto_estimate(account_sepolia_testnet):
    account = account_sepolia_testnet
    call = Call(
        to_addr=int(EMPTY_CONTRACT_ADDRESS_SEPOLIA, 0),
        selector=get_selector_from_name("empty"),
        calldata=[],
    )
    sign_invoke = await account.sign_invoke_v3(calls=call, auto_estimate=True)
    invoke = await account.client.send_transaction(sign_invoke)

    result = await account.client.wait_for_tx(tx_hash=invoke.transaction_hash)

    assert result.execution_status == TransactionExecutionStatus.SUCCEEDED


@pytest.mark.skip("TODO(#1621)")
@pytest.mark.asyncio
async def test_transaction_not_received_max_fee_too_small(account_sepolia_testnet):
    account = account_sepolia_testnet
    call = Call(
        to_addr=int(EMPTY_CONTRACT_ADDRESS_SEPOLIA, 0),
        selector=get_selector_from_name("empty"),
        calldata=[],
    )
    resource_bounds = ResourceBoundsMapping(
        l1_gas=ResourceBounds(max_amount=int(1e10), max_price_per_unit=int(1e10)),
        l2_gas=ResourceBounds(max_amount=int(1e10), max_price_per_unit=int(1e10)),
        l1_data_gas=ResourceBounds(max_amount=int(1e10), max_price_per_unit=int(1e10)),
    )
    sign_invoke = await account.sign_invoke_v3(
        calls=call, resource_bounds=resource_bounds
    )

    with pytest.raises(
        ClientError,
        match=r"Client failed with code 55. "
        r"Message: Account validation failed. Data: Max L1Gas price \(\d+\) is lower than the actual gas price: \d+.",
    ):
        await account.client.send_transaction(sign_invoke)


@pytest.mark.skip("TODO(#1621)")
@pytest.mark.asyncio
async def test_transaction_not_received_max_fee_too_big(account_sepolia_testnet):
    account = account_sepolia_testnet
    call = Call(
        to_addr=int(EMPTY_CONTRACT_ADDRESS_SEPOLIA, 0),
        selector=get_selector_from_name("empty"),
        calldata=[],
    )
    resource_bounds = ResourceBoundsMapping(
        l1_gas=ResourceBounds(max_amount=int(1e8), max_price_per_unit=int(1e15)),
        l2_gas=ResourceBounds(max_amount=int(1e14), max_price_per_unit=int(1e25)),
        l1_data_gas=ResourceBounds(max_amount=int(1e8), max_price_per_unit=int(1e15)),
    )
    sign_invoke = await account.sign_invoke_v3(
        calls=call, resource_bounds=resource_bounds
    )

    with pytest.raises(
        ClientError,
        match=r"Client failed with code 55. "
        r"Message: Account validation failed\. Data: Resources bounds \(\{.*\}\) exceed balance \(\d+\)\.",
    ):
        await account.client.send_transaction(sign_invoke)


@pytest.mark.asyncio
async def test_transaction_not_received_invalid_nonce(account_sepolia_testnet):
    account = account_sepolia_testnet
    call = Call(
        to_addr=int(EMPTY_CONTRACT_ADDRESS_SEPOLIA, 0),
        selector=get_selector_from_name("empty"),
        calldata=[],
    )
    sign_invoke = await account.sign_invoke_v3(
        calls=call, nonce=0, resource_bounds=MAX_RESOURCE_BOUNDS_SEPOLIA
    )

    with pytest.raises(ClientError, match=r".*nonce.*"):
        await account.client.send_transaction(sign_invoke)


@pytest.mark.asyncio
async def test_transaction_not_received_invalid_signature(account_sepolia_testnet):
    account = account_sepolia_testnet
    call = Call(
        to_addr=int(EMPTY_CONTRACT_ADDRESS_SEPOLIA, 0),
        selector=get_selector_from_name("empty"),
        calldata=[],
    )
    sign_invoke = await account.sign_invoke_v3(
        calls=call, resource_bounds=MAX_RESOURCE_BOUNDS_SEPOLIA
    )
    sign_invoke = dataclasses.replace(sign_invoke, signature=[0x21, 0x37])
    with pytest.raises(
        ClientError,
        match=r"Account validation failed",
    ) as exc:
        await account.client.send_transaction(sign_invoke)

    assert exc.value.data is not None
    assert "Data:" in exc.value.message


# TODO (#1219): move tests below to full_node_test.py
@pytest.mark.asyncio
async def test_estimate_message_fee(client_sepolia_testnet):
    client = client_sepolia_testnet
    # info about this message from
    # https://sepolia.starkscan.co/message-log/0x061e8c5211c705d0ab608e42f181edf4ef9ae891b3e568a6fe1c3b83076eefc2_0_1
    estimated_message = await client.estimate_message_fee(
        from_address="0x18e4a8e2badb5f5950758f46f8108e2c5d357b07",
        block_number=51569,
        to_address="0x054f677f3e952d023e2f31d74606270b676eaf493befbcfa2111f2b96a242362",
        entry_point_selector="0x03fa70707d0e831418fb142ca8fb7483611b84e89c0c42bf1fc2a7a5c40890ad",
        payload=[
            "0x1b0b3ddfc5264c441c9eee709011a863",
            "0xfbe265a54523fc9070e26bfc5aa145ab",
            "0x5469d9",
            "0x0",
        ],
    )

    assert isinstance(estimated_message, EstimatedFee)
    assert all(
        getattr(estimated_message, field.name) >= 0
        for field in dataclasses.fields(EstimatedFee)
        if isinstance(getattr(estimated_message, field.name), numbers.Number)
    )
    assert estimated_message.unit is not None


@pytest.mark.asyncio
async def test_estimate_message_fee_invalid_eth_address_assertion_error(
    client_sepolia_testnet,
):
    client = client_sepolia_testnet
    invalid_eth_address = "0xD"

    # info about this transaction from
    # https://sepolia.starkscan.co/tx/0x07041ce61b01f677ef05391ae1db043d0ea8b96309574ddcf90e7f59ec7d76d6
    with pytest.raises(
        AssertionError,
        match=f"Argument 'from_address': {invalid_eth_address} is not a valid Ethereum address.",
    ):
        _ = await client.estimate_message_fee(
            from_address=invalid_eth_address,
            block_number=51569,
            to_address="0x054f677f3e952d023e2f31d74606270b676eaf493befbcfa2111f2b96a242362",
            entry_point_selector="0x03fa70707d0e831418fb142ca8fb7483611b84e89c0c42bf1fc2a7a5c40890ad",
            payload=[
                "0x1b0b3ddfc5264c441c9eee709011a863",
                "0xfbe265a54523fc9070e26bfc5aa145ab",
                "0x5469d9",
                "0x0",
            ],
        )


@pytest.mark.parametrize(
    "from_address, to_address",
    (
        (
            "0xbe1259ff905cadbbaa62514388b71bdefb8aacc1",
            "0x1234",
        ),  # valid `from_address`, invalid `to_address`
        (
            "0xDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD",
            "0x06524771cb912945bf2db355b5a12355ca2e2ff05e15ee35366336a602293f2d",
        ),  # invalid `from_address` (passes through assert), valid `to_address`
    ),
)
@pytest.mark.asyncio
async def test_estimate_message_fee_throws(
    client_sepolia_testnet, from_address, to_address
):
    with pytest.raises(ClientError):
        _ = await client_sepolia_testnet.estimate_message_fee(
            block_number=80000,
            from_address=from_address,
            to_address=to_address,
            entry_point_selector="0x3248",
            payload=[
                "0x4359",
            ],
        )


@pytest.mark.asyncio
async def test_get_tx_receipt_reverted(client_sepolia_testnet):
    reverted_tx_hash = (
        "0x00fecca6a328dd11f40b79c30fe22d23bc6975d1a0923a95b90aff4016a84333"
    )

    res = await client_sepolia_testnet.get_transaction_receipt(tx_hash=reverted_tx_hash)

    assert res.execution_status == TransactionExecutionStatus.REVERTED
    assert res.finality_status == TransactionFinalityStatus.ACCEPTED_ON_L1
    assert "Got an exception while executing a hint" in res.revert_reason


@pytest.mark.parametrize(
    "block_number, index, expected_hash",
    [
        (81116, 0, 0x38FC01353196AEEBA62C74A8C8479FFF94AAA8CD4C3655782D49D755BBE63A8),
        (81116, 26, 0x3F873FE2CC884A88B8D4378EAC1786145F7167D61B0A9442DA15B0181582522),
        (80910, 23, 0x67C1E282F64DAD5682B1F377A5FDA1778311D894B2EE47A06058790A8B08460),
    ],
)
@pytest.mark.asyncio
async def test_get_transaction_by_block_id_and_index(
    client_sepolia_testnet, block_number, index, expected_hash
):
    tx = await client_sepolia_testnet.get_transaction_by_block_id(
        block_number=block_number, index=index
    )

    assert isinstance(tx, Transaction)
    assert tx.hash == expected_hash


@pytest.mark.asyncio
async def test_get_l1_message_hash(client_sepolia_testnet):
    tx_hash = "0x067d959200d65d4ad293aa4b0da21bb050a1f669bce37d215c6edbf041269c07"
    l1_message_hash = await client_sepolia_testnet.get_l1_message_hash(tx_hash)
    assert (
        hex(l1_message_hash)
        == "0x2e350fa9d830482605cb68be4fdb9f0cb3e1f95a0c51623ac1a5d1bd997c2090"
    )


@pytest.mark.asyncio
async def test_get_l1_message_hash_raises_on_incorrect_transaction_type(
    client_sepolia_testnet,
):
    tx_hash = "0x38FC01353196AEEBA62C74A8C8479FFF94AAA8CD4C3655782D49D755BBE63A8"
    with pytest.raises(
        TypeError, match=f"Transaction {tx_hash} is not a result of L1->L2 interaction."
    ):
        await client_sepolia_testnet.get_l1_message_hash(tx_hash)


@pytest.mark.asyncio
async def test_spec_version(client_sepolia_testnet):
    spec_version = await client_sepolia_testnet.spec_version()

    assert spec_version is not None
    assert isinstance(spec_version, str)


@pytest.mark.asyncio
async def test_get_transaction_status(client_sepolia_testnet):
    tx_status = await client_sepolia_testnet.get_transaction_status(
        tx_hash=0x06BF304EFEF9D0D28161C69A4660FA8AC769118A81FACE53BC8EA165BBB3F86F
    )

    assert tx_status.finality_status == TransactionStatus.ACCEPTED_ON_L1
    assert tx_status.execution_status == TransactionExecutionStatus.SUCCEEDED


@pytest.mark.asyncio
async def test_get_transaction_status_with_failure_reason(client_sepolia_testnet):
    # TODO(#1498): Potentially change tx for one that has a known failure reason
    # Originally, tx with hash 0x048d0e94d643f54f517271bd54936aa958d787c1b5d9d0a013ece6868ba9c8b7
    # has an unknown failure reason, therefore we need to mock it.
    with patch(
        f"{RpcHttpClient.__module__}.RpcHttpClient.call", AsyncMock()
    ) as mocked_tx_status_call_rpc:
        return_value = {
            "execution_status": "REVERTED",
            "finality_status": "ACCEPTED_ON_L2",
            "failure_reason": "Some failure reason",
        }
        mocked_tx_status_call_rpc.return_value = return_value

        tx_status = await client_sepolia_testnet.get_transaction_status(
            tx_hash=0x048D0E94D643F54F517271BD54936AA958D787C1B5D9D0A013ECE6868BA9C8B7
        )

        assert tx_status.finality_status == TransactionStatus.ACCEPTED_ON_L2
        assert tx_status.execution_status == TransactionExecutionStatus.REVERTED
        assert tx_status.failure_reason == "Some failure reason"


@pytest.mark.skip("TODO(#1621)")
@pytest.mark.asyncio
async def test_get_block_new_header_fields(client_sepolia_testnet):
    # testing l1_gas_price and starknet_version fields
    block = await client_sepolia_testnet.get_block_with_txs(block_number=155)

    assert block.starknet_version is not None
    assert block.l1_gas_price is not None
    assert block.l1_gas_price.price_in_wei > 0

    pre_confirmed_block = await client_sepolia_testnet.get_block_with_txs(
        block_number="pre_confirmed"
    )

    assert pre_confirmed_block.starknet_version is not None
    assert pre_confirmed_block.l1_gas_price is not None
    assert pre_confirmed_block.l1_gas_price.price_in_wei > 0


@pytest.mark.skip("TODO(#1621)")
@pytest.mark.asyncio
async def test_get_block_with_tx_hashes_new_header_fields(client_sepolia_testnet):
    # testing l1_gas_price and starknet_version fields
    block = await client_sepolia_testnet.get_block_with_tx_hashes(block_number=155)

    assert block.starknet_version is not None
    assert block.l1_gas_price is not None
    assert block.l1_gas_price.price_in_wei > 0

    pre_confirmed_block = await client_sepolia_testnet.get_block_with_tx_hashes(
        block_number="pre_confirmed"
    )

    assert pre_confirmed_block.starknet_version is not None
    assert pre_confirmed_block.l1_gas_price is not None
    assert pre_confirmed_block.l1_gas_price.price_in_wei > 0


@pytest.mark.parametrize(
    "tx_hash, tx_type",
    [
        (
            0x054270D103C875A613E013D1FD555EDCFF2085FECA9D7B4532243A8257FD5CF3,
            DeclareTransactionV3,
        ),
        (
            0x06718B783A0B888F5421C4EB76A532FEB9FD5167B2B09274298F79798C782B32,
            DeployAccountTransactionV3,
        ),
        (
            0x043868D939FA1B62B977FFFC659146688E954BBABEDA020CC99BAE1C220E4882,
            InvokeTransactionV3,
        ),
    ],
)
@pytest.mark.asyncio
async def test_get_transaction_v3(client_sepolia_testnet, tx_hash, tx_type):
    tx = await client_sepolia_testnet.get_transaction(tx_hash=tx_hash)
    assert isinstance(tx, tx_type)
    assert tx.version == 3
    assert isinstance(tx.resource_bounds, ResourceBoundsMapping)
    assert tx.paymaster_data == []
    assert tx.tip == 0
    assert tx.nonce_data_availability_mode == DAMode.L1
    assert tx.fee_data_availability_mode == DAMode.L1


@pytest.mark.asyncio
async def test_get_chain_id_sepolia_testnet(client_sepolia_testnet):
    chain_id = await client_sepolia_testnet.get_chain_id()
    assert isinstance(chain_id, str)
    assert chain_id == hex(StarknetChainId.SEPOLIA.value)


@pytest.mark.skip("TODO(#1659)")
@pytest.mark.asyncio
async def test_get_events_sepolia_testnet(client_sepolia_testnet):
    events_chunk = await client_sepolia_testnet.get_events(
        address=default_token_address_for_network(SEPOLIA),
        from_block_number=1000,
        to_block_number=1005,
        chunk_size=10,
    )
    assert isinstance(events_chunk, EventsChunk)
    assert len(events_chunk.events) == 10
    assert events_chunk.continuation_token is not None
    assert isinstance(events_chunk.events[0], EmittedEvent)
    assert events_chunk.events[0].block_number == 1000
    assert events_chunk.events[0].block_hash is not None
    assert events_chunk.events[0].from_address is not None
    assert events_chunk.events[0].data is not None
    assert events_chunk.events[0].keys is not None


@pytest.mark.asyncio
async def test_get_block_with_receipts(client_sepolia_testnet):
    block_with_receipts = await client_sepolia_testnet.get_block_with_receipts(
        block_number=48778
    )

    assert isinstance(block_with_receipts, StarknetBlockWithReceipts)
    assert block_with_receipts.status == BlockStatus.ACCEPTED_ON_L1
    assert len(block_with_receipts.transactions) == 43
    assert all(
        getattr(block_with_receipts, field.name) is not None
        for field in dataclasses.fields(BlockHeader)
    )


@pytest.mark.skip("TODO(#1621)")
@pytest.mark.asyncio
async def test_get_pre_confirmed_block_with_receipts(client_sepolia_testnet):
    block_with_receipts = await client_sepolia_testnet.get_block_with_receipts(
        block_number="pre_confirmed"
    )

    assert isinstance(block_with_receipts, PreConfirmedStarknetBlockWithReceipts)
    assert len(block_with_receipts.transactions) >= 0
    assert all(
        getattr(block_with_receipts, field.name) is not None
        for field in dataclasses.fields(PreConfirmedBlockHeader)
    )


@pytest.mark.asyncio
async def test_get_storage_proof(client_sepolia_testnet):
    # Nodes don't support storage proofs for blocks that are too far in the past, hence we need to get last block number
    block_number = await client_sepolia_testnet.get_block_number()
    storage_proof = await client_sepolia_testnet.get_storage_proof(
        block_number=block_number,
        contract_addresses=[int(STRK_FEE_CONTRACT_ADDRESS, 16)],
        contracts_storage_keys=[
            ContractsStorageKeys(
                contract_address=int(STRK_FEE_CONTRACT_ADDRESS, 16),
                storage_keys=["0x45524332305f62616c616e636573"],
            )
        ],
        class_hashes=[int(STRK_CLASS_HASH, 16)],
    )

    # We can't check exact values, as they change over the time
    assert len(storage_proof.classes_proof) > 0
    assert len(storage_proof.contracts_proof.nodes) > 0
    assert len(storage_proof.contracts_storage_proofs[0]) > 16


@pytest.mark.asyncio
async def test_get_compiled_casm(client_sepolia_testnet):
    compiled_casm = await client_sepolia_testnet.get_compiled_casm(
        class_hash=int(STRK_CLASS_HASH, 16)
    )

    assert isinstance(compiled_casm, CasmClass)
    assert len(compiled_casm.bytecode) == 20477
    assert len(compiled_casm.hints) == 931

    first_hint = compiled_casm.hints[0][1][0]
    assert isinstance(first_hint, TestLessThanOrEqual)
    assert first_hint.test_less_than_or_equal.dst.offset == 0
    assert first_hint.test_less_than_or_equal.dst.register == "AP"
    assert isinstance(first_hint.test_less_than_or_equal.lhs, Immediate)
    assert first_hint.test_less_than_or_equal.lhs.immediate == 0
    assert isinstance(first_hint.test_less_than_or_equal.rhs, Deref)
    assert first_hint.test_less_than_or_equal.rhs.deref.offset == -6
    assert first_hint.test_less_than_or_equal.rhs.deref.register == "FP"

    second_hint = compiled_casm.hints[1][1][0]
    assert isinstance(second_hint, TestLessThan)
    assert second_hint.test_less_than.dst.offset == 4
    assert second_hint.test_less_than.dst.register == "AP"
    assert isinstance(second_hint.test_less_than.lhs, Deref)
    assert second_hint.test_less_than.lhs.deref.offset == -1
    assert second_hint.test_less_than.lhs.deref.register == "AP"
    assert isinstance(second_hint.test_less_than.rhs, Immediate)
    assert (
        second_hint.test_less_than.rhs.immediate
        == 0x800000000000000000000000000000000000000000000000000000000000000
    )


@pytest.mark.asyncio
async def test_warning_on_incompatible_node_spec_version(client_sepolia_testnet):
    old_rpc_url = client_sepolia_testnet.url.replace("v0_9", "v0_8")
    node = FullNodeClient(old_rpc_url)

    pattern = (
        rf"RPC node with the url {old_rpc_url} uses incompatible version 0\.8\.1\. "
        rf"Expected version: {EXPECTED_RPC_VERSION}"
    )
    with pytest.warns(IncompatibleRPCVersionWarning, match=pattern):
        await node.get_chain_id()


@pytest.mark.asyncio
async def test_l1_accepted_block(account_sepolia_testnet):
    blk = await account_sepolia_testnet.client.get_block(block_number="l1_accepted")
    assert blk.block_hash
    assert blk.transactions is not None
    assert isinstance(blk, StarknetBlock)



from enum import IntEnum
from typing import Optional, Union

from starknet_py.common import int_from_bytes
from starknet_py.net.networks import MAINNET, SEPOLIA, SEPOLIA_INTEGRATION, Network


class StarknetChainId(IntEnum):
    """
    An enumeration representing Starknet chain IDs.
    """

    MAINNET = int_from_bytes(b"SN_MAIN")
    SEPOLIA = int_from_bytes(b"SN_SEPOLIA")
    SEPOLIA_INTEGRATION = int_from_bytes(b"SN_INTEGRATION_SEPOLIA")


RECOGNIZED_CHAIN_IDS = [
    StarknetChainId.MAINNET,
    StarknetChainId.SEPOLIA,
    StarknetChainId.SEPOLIA_INTEGRATION,
]


def chain_from_network(
    net: Network, chain: Optional[StarknetChainId] = None
) -> StarknetChainId:
    mapping = {
        MAINNET: StarknetChainId.MAINNET,
        SEPOLIA: StarknetChainId.SEPOLIA,
        SEPOLIA_INTEGRATION: StarknetChainId.SEPOLIA_INTEGRATION,
    }

    if isinstance(net, str) and net in mapping:
        return mapping[net]

    if not chain:
        raise ValueError("Chain is required when not using predefined networks.")

    return chain


ChainId = Union[StarknetChainId, int]
Chain = Union[str, ChainId]


def parse_chain(chain: Chain) -> ChainId:
    if isinstance(chain, str):
        try:
            return int(chain, 16)
        except ValueError:
            return int_from_bytes(chain.encode())
    else:
        return chain



# pylint: disable=redefined-outer-name
from typing import AsyncGenerator, Dict, List, Tuple

import pytest
import pytest_asyncio

from starknet_py.common import create_sierra_compiled_contract
from starknet_py.contract import Contract
from starknet_py.hash.selector import get_selector_from_name
from starknet_py.net.account.account import Account
from starknet_py.net.account.base_account import BaseAccount
from starknet_py.tests.e2e.client.fixtures.prepare_net_for_gateway_test import (
    PreparedNetworkData,
    prepare_net_for_tests,
)
from starknet_py.tests.e2e.fixtures.accounts import AccountToBeDeployedDetailsFactory
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS
from starknet_py.tests.e2e.fixtures.contracts_v1 import declare_contract
from starknet_py.tests.e2e.fixtures.misc import load_contract


@pytest_asyncio.fixture(scope="package")
async def balance_class_and_transaction_hash(account: BaseAccount) -> Tuple[int, int]:
    contract = load_contract("Balance")
    class_hash, transaction_hash = await declare_contract(
        account,
        contract["sierra"],
        contract["casm"],
    )
    return class_hash, transaction_hash


@pytest_asyncio.fixture(scope="package")
async def deployed_balance_contract(
    account: BaseAccount,
    balance_class_and_transaction_hash,
    balance_abi,
) -> Contract:
    class_hash, _ = balance_class_and_transaction_hash

    deploy_result = await Contract.deploy_contract_v3(
        account=account,
        abi=balance_abi,
        class_hash=class_hash,
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )
    await deploy_result.wait_for_acceptance()

    return deploy_result.deployed_contract


@pytest_asyncio.fixture(scope="package")
async def deployed_balance_contract_2(
    account: BaseAccount,
    balance_class_and_transaction_hash,
    balance_abi,
) -> Contract:
    class_hash, _ = balance_class_and_transaction_hash
    deploy_result = await Contract.deploy_contract_v3(
        account=account,
        abi=balance_abi,
        class_hash=class_hash,
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )
    await deploy_result.wait_for_acceptance()

    return deploy_result.deployed_contract


@pytest.fixture(scope="package")
def balance_abi() -> List:
    compiled_contract = create_sierra_compiled_contract(
        compiled_contract=load_contract("Balance")["sierra"]
    )
    assert compiled_contract.parsed_abi is not None
    return compiled_contract.parsed_abi


@pytest.fixture()
def block_with_invoke_number(prepare_network: Tuple[str, PreparedNetworkData]) -> int:
    """
    Returns number of the block with invoke transaction
    """
    _, prepared_data = prepare_network
    return prepared_data.block_with_invoke_number


@pytest.fixture()
def block_with_declare_number(prepare_network: Tuple[str, PreparedNetworkData]) -> int:
    """
    Returns number of the block with declare transaction
    """
    _, prepared_data = prepare_network
    return prepared_data.block_with_declare_number


@pytest.fixture()
def block_with_declare_hash(prepare_network: Tuple[str, PreparedNetworkData]) -> int:
    """
    Returns hash of the block with declare transaction
    """
    _, prepared_data = prepare_network
    return prepared_data.block_with_declare_hash


@pytest.fixture()
def invoke_transaction(prepare_network: Tuple[str, PreparedNetworkData]) -> Dict:
    """
    Returns basic data of Invoke
    """
    _, prepared_data = prepare_network
    return {
        "hash": prepared_data.invoke_transaction_hash,
        "calldata": [1234],
        "entry_point_selector": get_selector_from_name("increase_balance"),
    }


@pytest.fixture()
def invoke_transaction_hash(invoke_transaction: Dict) -> int:
    """
    Returns hash of Invoke
    """
    return invoke_transaction["hash"]


@pytest.fixture()
def invoke_transaction_calldata(invoke_transaction: Dict) -> int:
    """
    Returns calldata of Invoke
    """
    return invoke_transaction["calldata"]


@pytest.fixture()
def invoke_transaction_selector(invoke_transaction: Dict) -> int:
    """
    Returns entry_point_selector of Invoke
    """
    return invoke_transaction["entry_point_selector"]


@pytest.fixture()
def declare_transaction_hash(prepare_network: Tuple[str, PreparedNetworkData]) -> int:
    """
    Returns hash of the DeclareTransaction
    """
    _, prepared_data = prepare_network
    return prepared_data.declare_transaction_hash


@pytest.fixture()
def contract_address(prepare_network: Tuple[str, PreparedNetworkData]) -> int:
    """
    Returns an address of the deployed contract
    """
    _, prepared_data = prepare_network
    return prepared_data.contract_address


# `contract_address` was used in other tests, which modified its storage values. This overlap
# caused test interdependencies, leading to inconsistent results in `test_get_storage_at`
# and `test_call_contract`, hence the introduction of `contract_address_2`.
@pytest.fixture()
def contract_address_2(prepare_network: Tuple[str, PreparedNetworkData]) -> int:
    """
    Returns an address of the deployed contract
    """
    _, prepared_data = prepare_network
    return prepared_data.contract_address_2


@pytest.fixture()
def class_hash(prepare_network: Tuple[str, PreparedNetworkData]) -> int:
    """
    Returns class hash of the deployed contract
    """
    _, prepared_data = prepare_network
    return prepared_data.class_hash


@pytest_asyncio.fixture(scope="package")
async def prepare_network(
    devnet,
    account: Account,
    deploy_account_details_factory: AccountToBeDeployedDetailsFactory,
    balance_class_and_transaction_hash: Tuple[int, int],
    deployed_balance_contract: Contract,
    deployed_balance_contract_2: Contract,
) -> AsyncGenerator[Tuple[str, PreparedNetworkData], None]:
    """
    Adds transactions to the network. Returns network address and PreparedNetworkData
    """
    net = devnet
    class_hash, transaction_hash = balance_class_and_transaction_hash
    details = await deploy_account_details_factory.get()

    prepared_data = await prepare_net_for_tests(
        account,
        deploy_account_details=details,
        transaction_hash=transaction_hash,
        contract=deployed_balance_contract,
        contract_2=deployed_balance_contract_2,
        declare_class_hash=class_hash,
    )

    yield net, prepared_data



from unittest.mock import AsyncMock, patch

import pytest

from starknet_py.contract import Contract
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS
from starknet_py.tests.e2e.fixtures.misc import ContractVersion, load_contract


@pytest.mark.asyncio
async def test_throws_when_cairo1_without_compiled_contract_casm_and_class_hash(
    account,
):
    error_message = (
        "For Cairo 1.0 contracts, either the 'compiled_class_hash' or the 'compiled_contract_casm' "
        "argument must be provided."
    )
    compiled_contract = load_contract("Map")["sierra"]

    with pytest.raises(ValueError, match=error_message):
        await Contract.declare_v3(
            account,
            compiled_contract=compiled_contract,
            resource_bounds=MAX_RESOURCE_BOUNDS,
        )

    with pytest.raises(ValueError, match=error_message):
        await Contract.declare_v3(
            account,
            compiled_contract=compiled_contract,
            resource_bounds=MAX_RESOURCE_BOUNDS,
        )


@pytest.mark.asyncio
async def test_declare_v3(
    account,
):
    contract = load_contract(contract_name="TestContract", version=ContractVersion.V2)

    tip = 12345
    declare_result = await Contract.declare_v3(
        account,
        compiled_contract=contract["sierra"],
        compiled_contract_casm=contract["casm"],
        resource_bounds=MAX_RESOURCE_BOUNDS,
        tip=tip,
    )

    await declare_result.wait_for_acceptance()
    assert declare_result.declare_transaction.tip == tip


@pytest.mark.asyncio
async def test_declare_v3_auto_estimate_tip(
    account,
    get_block_with_txs_path,
    block_with_tips_mock,
):
    contract = load_contract(contract_name="TestContract2", version=ContractVersion.V2)

    with patch(get_block_with_txs_path, AsyncMock()) as get_block_with_txs_mock:
        get_block_with_txs_mock.return_value = block_with_tips_mock

        declare_result = await Contract.declare_v3(
            account,
            compiled_contract=contract["sierra"],
            compiled_contract_casm=contract["casm"],
            resource_bounds=MAX_RESOURCE_BOUNDS,
            auto_estimate_tip=True,
        )

    await declare_result.wait_for_acceptance()
    assert declare_result.declare_transaction.tip == 2



import pytest

from starknet_py.net.client_models import (
    DeclareTransactionTrace,
    DeclareTransactionV3,
    DeployAccountTransactionTrace,
    DeployAccountTransactionV3,
    InvokeTransactionTrace,
    InvokeTransactionV3,
    L1HandlerTransaction,
    L1HandlerTransactionTrace,
    RevertedFunctionInvocation,
)


@pytest.mark.asyncio
@pytest.mark.skip("TODO(#1562)")
async def test_trace_transaction_invoke_v3(client_sepolia_testnet):
    invoke_tx_hash = 0x26476DA48E56E5E7025543AD0BB9105DF00EE08571C6D17C4207462FF7717C4
    trace = await client_sepolia_testnet.trace_transaction(tx_hash=invoke_tx_hash)
    tx = await client_sepolia_testnet.get_transaction(tx_hash=invoke_tx_hash)

    assert isinstance(tx, InvokeTransactionV3)
    assert isinstance(trace, InvokeTransactionTrace)
    assert trace.execute_invocation is not None
    assert trace.execution_resources is not None


@pytest.mark.asyncio
@pytest.mark.skip("TODO(#1562)")
async def test_trace_transaction_declare_v3(client_sepolia_testnet):
    declare_tx_hash = 0x6054540622D534FFFFB162A0E80C21BC106581EAFEB3EFAD29385B78E04983D
    trace = await client_sepolia_testnet.trace_transaction(tx_hash=declare_tx_hash)
    tx = await client_sepolia_testnet.get_transaction(tx_hash=declare_tx_hash)

    assert isinstance(tx, DeclareTransactionV3)
    assert isinstance(trace, DeclareTransactionTrace)
    assert trace.execution_resources is not None


@pytest.mark.asyncio
@pytest.mark.skip("TODO(#1562)")
async def test_trace_transaction_deploy_account_v3(client_sepolia_testnet):
    deploy_account_tx_hash = (
        0x06718B783A0B888F5421C4EB76A532FEB9FD5167B2B09274298F79798C782B32
    )
    trace = await client_sepolia_testnet.trace_transaction(
        tx_hash=deploy_account_tx_hash
    )
    tx = await client_sepolia_testnet.get_transaction(tx_hash=deploy_account_tx_hash)

    assert isinstance(tx, DeployAccountTransactionV3)
    assert isinstance(trace, DeployAccountTransactionTrace)
    assert trace.constructor_invocation is not None
    assert trace.execution_resources is not None


@pytest.mark.asyncio
@pytest.mark.skip("TODO(#1562)")
async def test_trace_transaction_l1_handler(client_sepolia_testnet):
    l1_handler_tx_hash = (
        0x4C8C57B3AB646EF56AEF3DEF69A01BC86D049B98F25EBFE3699334D86C24D5
    )
    trace = await client_sepolia_testnet.trace_transaction(tx_hash=l1_handler_tx_hash)
    tx = await client_sepolia_testnet.get_transaction(tx_hash=l1_handler_tx_hash)

    assert isinstance(tx, L1HandlerTransaction)
    assert isinstance(trace, L1HandlerTransactionTrace)
    assert trace.function_invocation is not None
    assert trace.execution_resources is not None


@pytest.mark.asyncio
@pytest.mark.skip("TODO(#1562)")
async def test_trace_transaction_reverted(client_sepolia_testnet):
    tx_hash = 0x00FECCA6A328DD11F40B79C30FE22D23BC6975D1A0923A95B90AFF4016A84333
    trace = await client_sepolia_testnet.trace_transaction(tx_hash=tx_hash)

    assert isinstance(trace.execute_invocation, RevertedFunctionInvocation)


@pytest.mark.asyncio
@pytest.mark.skip("TODO(#1562)")
async def test_get_block_traces(client_sepolia_testnet):
    block_number = 80000
    block_transaction_traces = await client_sepolia_testnet.trace_block_transactions(
        block_number=block_number
    )
    block = await client_sepolia_testnet.get_block(block_number=block_number)

    assert len(block_transaction_traces) == len(block.transactions)
    for i, block_transaction_trace in enumerate(block_transaction_traces):
        assert block_transaction_trace.transaction_hash == block.transactions[i].hash



from dataclasses import dataclass
from typing import Dict, Generator, OrderedDict, Tuple, Union

from starknet_py.serialization._context import (
    DeserializationContext,
    SerializationContext,
)
from starknet_py.serialization.data_serializers.cairo_data_serializer import (
    CairoDataSerializer,
)
from starknet_py.serialization.tuple_dataclass import TupleDataclass


@dataclass
class EnumSerializer(CairoDataSerializer[Union[Dict, TupleDataclass], TupleDataclass]):
    """
    Serializer of enums.
    Can serialize a dictionary and TupleDataclass.
    Deserializes data to a TupleDataclass.

    Example:
        enum MyEnum {
          a: u128,
          b: u128
        }

        {"a": 1} => [0, 1]
        {"b": 100} => [1, 100]
        TupleDataclass(variant='a', value=100) => [0, 100]
    """

    serializers: OrderedDict[str, CairoDataSerializer]

    def deserialize_with_context(
        self, context: DeserializationContext
    ) -> TupleDataclass:
        [variant_index] = context.reader.read(1)
        variant_name, serializer = self._get_variant(variant_index)

        with context.push_entity("enum.variant: " + variant_name):
            result_dict = {
                "variant": variant_name,
                "value": serializer.deserialize_with_context(context),
            }

        return TupleDataclass.from_dict(result_dict)

    def serialize_with_context(
        self, context: SerializationContext, value: Union[Dict, TupleDataclass]
    ) -> Generator[int, None, None]:
        if isinstance(value, Dict):
            items = list(value.items())
            if len(items) != 1:
                raise ValueError(
                    "Can serialize only one enum variant, got: " + str(len(items))
                )

            variant_name, variant_value = items[0]
        else:
            variant_name, variant_value = value

        yield self._get_variant_index(variant_name)
        yield from self.serializers[variant_name].serialize_with_context(
            context, variant_value
        )

    def _get_variant(self, variant_index: int) -> Tuple[str, CairoDataSerializer]:
        return list(self.serializers.items())[variant_index]

    def _get_variant_index(self, variant_name: str) -> int:
        return list(self.serializers.keys()).index(variant_name)



import pytest

from starknet_py.hash.transaction import (
    CommonTransactionV3Fields,
    TransactionHashPrefix,
    compute_declare_v2_transaction_hash,
    compute_declare_v3_transaction_hash,
    compute_deploy_account_transaction_hash,
    compute_deploy_account_v3_transaction_hash,
    compute_invoke_transaction_hash,
    compute_invoke_v3_transaction_hash,
    compute_transaction_hash,
)
from starknet_py.net.client_models import DAMode
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS


@pytest.mark.parametrize(
    "data, expected_hash",
    (
        (
            [TransactionHashPrefix.INVOKE, 2, 3, 4, [5], 6, 7, [8]],
            1176541524183318710439453346174486593835028723304878846501762065493748312757,
        ),
        (
            [TransactionHashPrefix.L1_HANDLER, 15, 39, 74, [74], 39, 15, [28]],
            1226653506056503634668815848352741482067480791322607584496401451909331743178,
        ),
        (
            [
                TransactionHashPrefix.INVOKE,
                0x0,
                0x2A,
                0x64,
                [],
                0x0,
                0x534E5F474F45524C49,
            ],
            0x7D260744DE9D8C55E7675A34512D1951A7B262C79E685D26599EDD2948DE959,
        ),
    ),
)
def test_compute_transaction_hash(data, expected_hash):
    assert compute_transaction_hash(*data) == expected_hash


@pytest.mark.parametrize(
    "data, expected_hash",
    (
        (
            {
                "contract_address": 2,
                "class_hash": 3,
                "constructor_calldata": [4],
                "salt": 5,
                "version": 6,
                "max_fee": 7,
                "chain_id": 8,
                "nonce": 9,
            },
            0x6199E956E541CBB06589C4A63C2578A8ED6B697C0FA35B002F48923DFE648EE,
        ),
    ),
)
def test_compute_deploy_account_transaction_hash(data, expected_hash):
    assert compute_deploy_account_transaction_hash(**data) == expected_hash


@pytest.mark.parametrize(
    "data, expected_hash",
    (
        (
            {
                "class_hash": 2,
                "sender_address": 3,
                "version": 4,
                "max_fee": 5,
                "chain_id": 6,
                "nonce": 7,
                "compiled_class_hash": 8,
            },
            0x67EA411072DD2EF3BA36D9680F040A02E599F80F4770E204ECBB2C47C226793,
        ),
    ),
)
def test_compute_declare_v2_transaction_hash(data, expected_hash):
    assert compute_declare_v2_transaction_hash(**data) == expected_hash


@pytest.mark.parametrize(
    "data, expected_hash",
    (
        (
            {
                "sender_address": 3,
                "version": 4,
                "calldata": [5],
                "max_fee": 6,
                "chain_id": 7,
                "nonce": 8,
            },
            0x505BBF7CD810531C53526631078DAA314BFD036C80C7C6E3A02C608DB8E31DE,
        ),
    ),
)
def test_compute_invoke_transaction_hash(data, expected_hash):
    assert compute_invoke_transaction_hash(**data) == expected_hash


@pytest.mark.parametrize(
    "common_data, declare_data, expected_hash",
    (
        (
            {
                "address": 0x52125C1E043126C637D1436D9551EF6C4F6E3E36945676BBD716A56E3A41B7A,
                "chain_id": 0x534E5F474F45524C49,
                "nonce": 0x675,
                "tip": 0x0,
                "paymaster_data": [],
                "nonce_data_availability_mode": DAMode.L1,
                "fee_data_availability_mode": DAMode.L1,
                "tx_prefix": TransactionHashPrefix.DECLARE,
                "version": 0x3,
            },
            {
                "class_hash": 0x2338634F11772EA342365ABD5BE9D9DC8A6F44F159AD782FDEBD3DB5D969738,
                "compiled_class_hash": 0x17B5169C770D0E49100AB0FC672A49CA90CC572F21F79A640B5227B19D3A447,
                "account_deployment_data": [],
            },
            0x16081C54C3BEDC5079E0024896BFD85ED7E57FFD52B138CBC73AF0F34C7FCCE,
        ),
    ),
)
def test_compute_declare_v3_transaction_hash(common_data, declare_data, expected_hash):
    assert (
        compute_declare_v3_transaction_hash(
            **declare_data,
            common_fields=CommonTransactionV3Fields(
                **common_data, resource_bounds=MAX_RESOURCE_BOUNDS
            ),
        )
        == expected_hash
    )


@pytest.mark.parametrize(
    "common_data, invoke_data, expected_hash",
    (
        (
            {
                "address": 0x35ACD6DD6C5045D18CA6D0192AF46B335A5402C02D41F46E4E77EA2C951D9A3,
                "chain_id": 0x534E5F474F45524C49,
                "nonce": 0x5,
                "tip": 0x0,
                "paymaster_data": [],
                "nonce_data_availability_mode": DAMode.L1,
                "fee_data_availability_mode": DAMode.L1,
                "tx_prefix": TransactionHashPrefix.INVOKE,
                "version": 0x3,
            },
            {
                "calldata": [
                    0x2,
                    0x6359ED638DF79B82F2F9DBF92ABBCB41B57F9DD91EAD86B1C85D2DEE192C,
                    0xB17D8A2731BA7CA1816631E6BE14F0FC1B8390422D649FA27F0FBB0C91EEA8,
                    0x0,
                    0x3FE8E4571772BBE0065E271686BD655EFD1365A5D6858981E582F82F2C10313,
                    0x2FD9126EE011F3A837CEA02E32AE4EE73342D827E216998E5616BAB88D8B7EA,
                    0x1,
                    0x2FD9126EE011F3A837CEA02E32AE4EE73342D827E216998E5616BAB88D8B7EA,
                ],
                "account_deployment_data": [],
            },
            0x119386B4AAAEF905BF027D3DD2734474C5E944942BF3FBD8FDB442704D32B8B,
        ),
    ),
)
def test_compute_invoke_v3_transaction_hash(common_data, invoke_data, expected_hash):
    assert (
        compute_invoke_v3_transaction_hash(
            **invoke_data,
            common_fields=CommonTransactionV3Fields(
                **common_data, resource_bounds=MAX_RESOURCE_BOUNDS
            ),
        )
        == expected_hash
    )


# TODO(#1498): Remove the skip mark
@pytest.mark.skip
@pytest.mark.parametrize(
    "common_data, deploy_account_data, expected_hash",
    (
        (
            {
                "address": 0x2FAB82E4AEF1D8664874E1F194951856D48463C3E6BF9A8C68E234A629A6F50,
                "chain_id": 0x534E5F474F45524C49,
                "nonce": 0x0,
                "tip": 0x0,
                "paymaster_data": [],
                "nonce_data_availability_mode": DAMode.L1,
                "fee_data_availability_mode": DAMode.L1,
                "tx_prefix": TransactionHashPrefix.DEPLOY_ACCOUNT,
                "version": 0x3,
            },
            {
                "constructor_calldata": [
                    0x5CD65F3D7DAEA6C63939D659B8473EA0C5CD81576035A4D34E52FB06840196C
                ],
                "contract_address_salt": 0x0,
                "class_hash": 0x2338634F11772EA342365ABD5BE9D9DC8A6F44F159AD782FDEBD3DB5D969738,
            },
            0x29FD7881F14380842414CDFDD8D6C0B1F2174F8916EDCFEB1EDE1EB26AC3EF0,
        ),
    ),
)
def test_compute_deploy_account_v3_transaction_hash(
    common_data, deploy_account_data, expected_hash
):
    assert (
        compute_deploy_account_v3_transaction_hash(
            **deploy_account_data,
            common_fields=CommonTransactionV3Fields(
                **common_data,
                resource_bounds=MAX_RESOURCE_BOUNDS,
            ),
        )
        == expected_hash
    )



from typing import List

from eth_keys import keys  # pyright: ignore[reportPrivateImportUsage]
from eth_keys.datatypes import Signature

from starknet_py.net.client_models import Hash
from starknet_py.net.models import AccountTransaction
from starknet_py.net.models.chains import ChainId
from starknet_py.net.signer.base_signer import BaseSigner
from starknet_py.serialization import Uint256Serializer
from starknet_py.utils.typed_data import TypedData


class EthSigner(BaseSigner):
    def __init__(
        self,
        private_key: Hash,
        chain_id: ChainId,
    ):
        if isinstance(private_key, str):
            private_key = int(private_key, 0)

        private_key_bytes = private_key.to_bytes(32, byteorder="big")
        self._private_key = keys.PrivateKey(private_key_bytes)
        self.chain_id = chain_id

    @property
    def public_key(self) -> int:
        return int.from_bytes(self._private_key.public_key.to_bytes(), byteorder="big")

    def sign_message(self, typed_data: TypedData, account_address: int) -> List[int]:
        msg_hash = typed_data.message_hash(account_address)
        signature = self._private_key.sign_msg_hash(
            msg_hash.to_bytes(32, byteorder="big")
        )

        return _serialize_signature(signature)

    def sign_transaction(self, transaction: AccountTransaction) -> List[int]:
        tx_hash = transaction.calculate_hash(self.chain_id)
        signature = self._private_key.sign_msg_hash(
            tx_hash.to_bytes(32, byteorder="big")
        )

        return _serialize_signature(signature)


# eth signature components `r` and `s` are 32 bytes each, so they are stored as u256
def _serialize_signature(signature: Signature) -> List[int]:
    serializer = Uint256Serializer()
    return (
        serializer.serialize(signature.r)
        + serializer.serialize(signature.s)
        + [signature.v]
    )



# pylint: disable=unused-variable
import pytest
from aiohttp import ClientSession

from starknet_py.contract import Contract
from starknet_py.hash.selector import get_selector_from_name
from starknet_py.hash.storage import get_storage_var_address
from starknet_py.net.client_models import Call
from starknet_py.net.full_node_client import FullNodeClient
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS


def test_init():
    # docs-start: init
    client = FullNodeClient(node_url="https://your.node.url")
    # docs-end: init


@pytest.mark.asyncio
async def test_init_with_custom_client_session():
    # docs-start: init
    # or with custom client session
    session = ClientSession()
    client = FullNodeClient(node_url="https://your.node.url", session=session)
    # perform operations...
    # close the session
    await session.close()
    # docs-end: init


@pytest.mark.asyncio
async def test_get_block(client, block_with_declare_hash):
    # docs-start: get_block
    block = await client.get_block(block_number="latest")
    block = await client.get_block(block_number=0)
    # or
    block_hash = "0x0"
    # docs-end: get_block
    block_hash = block_with_declare_hash
    # docs-start: get_block
    block = await client.get_block(block_hash=block_hash)
    # docs-end: get_block


@pytest.mark.skip(reason="TODO(#1659)")
@pytest.mark.asyncio
async def test_get_state_update(
    client, declare_transaction_hash
):  # pylint: disable=unused-argument
    # parameter `declare_transaction_hash` is needed because devnet blockchain is empty without it
    # and methods return invalid data

    # docs-start: get_state_update
    state_update = await client.get_state_update(block_number="latest")
    state_update = await client.get_state_update(block_number=1)
    # or
    block_hash = "0x0"
    # docs-end: get_state_update
    block_hash = state_update.block_hash
    # docs-start: get_state_update
    state_update = await client.get_state_update(block_hash=block_hash)
    # docs-end: get_state_update


@pytest.mark.asyncio
async def test_get_storage_at(client, map_contract):
    address = map_contract.address
    # docs-start: get_storage_at
    storage_value = await client.get_storage_at(
        contract_address=address,
        key=get_storage_var_address("storage_var name"),
        block_number="latest",
    )
    # docs-end: get_storage_at


@pytest.mark.asyncio
async def test_get_transaction(client, declare_transaction_hash):
    # docs-start: get_transaction
    transaction_hash = 0x1 or 1 or "0x1"
    # docs-end: get_transaction
    transaction_hash = declare_transaction_hash
    # docs-start: get_transaction
    transaction = await client.get_transaction(tx_hash=transaction_hash)
    # docs-end: get_transaction


@pytest.mark.asyncio
async def test_get_transaction_receipt(client, declare_transaction_hash):
    transaction_hash = declare_transaction_hash
    # docs-start: get_transaction_receipt
    transaction_receipt = await client.get_transaction_receipt(tx_hash=transaction_hash)
    # docs-end: get_transaction_receipt


@pytest.mark.asyncio
async def test_estimate_fee(client, deploy_account_transaction):
    transaction = deploy_account_transaction
    # docs-start: estimate_fee
    # a single transaction
    estimated_fee = await client.estimate_fee(tx=transaction)
    # or a list of transactions
    estimated_fee = await client.estimate_fee(tx=[transaction])
    # docs-end: estimate_fee


@pytest.mark.asyncio
async def test_call_contract(client, contract_address):
    # docs-start: call_contract
    response = await client.call_contract(
        call=Call(
            to_addr=contract_address,
            selector=get_selector_from_name("increase_balance"),
            calldata=[123],
        ),
        block_number="latest",
    )
    # docs-end: call_contract


@pytest.mark.asyncio
async def test_get_class_hash_at(client, contract_address):
    # docs-start: get_class_hash_at
    address = 0x1 or 1 or "0x1"
    # docs-end: get_class_hash_at
    address = contract_address
    # docs-start: get_class_hash_at
    class_hash = await client.get_class_hash_at(
        contract_address=address, block_number="latest"
    )
    # docs-end: get_class_hash_at


@pytest.mark.asyncio
async def test_get_class_by_hash(client, class_hash):
    # docs-start: get_class_by_hash
    hash_ = 0x1 or 1 or "0x1"
    # docs-end: get_class_by_hash
    hash_ = class_hash
    # docs-start: get_class_by_hash
    contract_class = await client.get_class_by_hash(class_hash=hash_)
    # docs-end: get_class_by_hash


@pytest.mark.skip("TODO(#1560)")
@pytest.mark.asyncio
async def test_get_transaction_by_block_id(client):
    # docs-start: get_transaction_by_block_id
    transaction = await client.get_transaction_by_block_id(
        index=0, block_number="latest"
    )
    # docs-end: get_transaction_by_block_id


@pytest.mark.asyncio
async def test_get_block_transaction_count(client):
    # docs-start: get_block_transaction_count
    num_of_transactions = await client.get_block_transaction_count(
        block_number="latest"
    )
    # docs-end: get_block_transaction_count


@pytest.mark.asyncio
async def test_get_class_at(client, contract_address):
    # docs-start: get_class_at
    address = 0x1 or 1 or "0x1"
    # docs-end: get_class_at
    address = contract_address
    # docs-start: get_class_at
    contract_class = await client.get_class_at(
        contract_address=address, block_number="latest"
    )
    # docs-end: get_class_at


@pytest.mark.asyncio
async def test_get_contract_nonce(client, contract_address):
    # docs-start: get_contract_nonce
    address = 0x1 or 1 or "0x1"
    # docs-end: get_contract_nonce
    address = contract_address
    # docs-start: get_contract_nonce
    nonce = await client.get_contract_nonce(
        contract_address=address, block_number="latest"
    )
    # docs-end: get_contract_nonce


@pytest.mark.asyncio
async def test_get_events(client, contract_address):
    # docs-start: get_events
    address = 0x1 or 1 or "0x1"
    # docs-end: get_events
    address = contract_address
    # docs-start: get_events
    events_response = await client.get_events(
        address=address,
        keys=[[1, 2], [], [3]],
        from_block_number=0,
        to_block_number="latest",
        follow_continuation_token=True,
        chunk_size=47,
    )
    # docs-end: get_events


# TODO (#1219): fix that after update to RPC to v0.6.0
@pytest.mark.xfail(
    reason="Passing devnet client without implemented methods - test simply for a code example."
)
@pytest.mark.asyncio
async def test_trace_block_transactions(client):
    # docs-start: trace_block_transactions
    block_number = 800002
    block_transaction_traces = await client.trace_block_transactions(
        block_number=block_number
    )
    # docs-end: trace_block_transactions


# TODO (#1219): fix that after update to RPC to v0.6.0
@pytest.mark.xfail(
    reason="Passing devnet client without implemented methods - test simply for a code example."
)
@pytest.mark.asyncio
async def test_trace_transaction(client):
    # docs-start: trace_transaction
    transaction_hash = "0x123"
    # docs-end: trace_transaction
    transaction_hash = (
        "0x31e9adddefb28fab4d2ef9a6907e5805f5f793f5198618119a5347e6fc4af57"
    )
    # docs-start: trace_transaction
    transaction_trace = await client.trace_transaction(tx_hash=transaction_hash)
    # docs-end: trace_transaction


@pytest.mark.asyncio
@pytest.mark.skip(
    "TODO(#1560): There is no `l1_data_gas` in `execution_resources` from devnet"
)
async def test_simulate_transactions(
    account, deployed_balance_contract, deploy_account_transaction
):
    assert isinstance(deployed_balance_contract, Contract)
    contract_address = deployed_balance_contract.address
    second_transaction = deploy_account_transaction
    # docs-start: simulate_transactions
    call = Call(
        to_addr=contract_address,
        selector=get_selector_from_name("method_name"),
        calldata=[0xCA11DA7A],
    )
    first_transaction = await account.sign_invoke_v3(
        calls=call, resource_bounds=MAX_RESOURCE_BOUNDS
    )
    # docs-end: simulate_transactions

    call = Call(
        to_addr=deployed_balance_contract.address,
        selector=get_selector_from_name("increase_balance"),
        calldata=[0x10],
    )
    first_transaction = await account.sign_invoke_v3(calls=call, auto_estimate=True)

    # docs-start: simulate_transactions
    # one transaction
    simulated_txs = await account.client.simulate_transactions(
        transactions=[first_transaction], block_number="latest"
    )
    # or multiple
    simulated_txs = await account.client.simulate_transactions(
        transactions=[first_transaction, second_transaction], block_number="latest"
    )
    # docs-end: simulate_transactions



import sys
from typing import cast
from unittest.mock import AsyncMock, patch

import pytest

from starknet_py.common import create_casm_class
from starknet_py.constants import ARGENT_V040_CLASS_HASH
from starknet_py.hash.address import compute_address
from starknet_py.hash.casm_class_hash import compute_casm_class_hash
from starknet_py.hash.selector import get_selector_from_name
from starknet_py.net.account.account import Account
from starknet_py.net.account.base_account import BaseAccount
from starknet_py.net.client_errors import ClientError
from starknet_py.net.client_models import (
    Call,
    DeclareTransactionV3,
    DeployAccountTransactionResponse,
    DeployAccountTransactionV3,
    EstimatedFee,
    InvokeTransactionV3,
    PriceUnit,
    ResourceBounds,
    ResourceBoundsMapping,
    SierraContractClass,
    TransactionExecutionStatus,
    TransactionFinalityStatus,
)
from starknet_py.net.full_node_client import FullNodeClient
from starknet_py.net.models import StarknetChainId
from starknet_py.net.models.transaction import DeclareV3, DeployAccountV3, InvokeV3
from starknet_py.net.signer.key_pair import KeyPair
from starknet_py.net.udc_deployer.deployer import Deployer
from starknet_py.tests.e2e.fixtures.accounts import AccountPrerequisites
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS
from starknet_py.tests.e2e.fixtures.misc import load_contract


@pytest.mark.run_on_devnet
@pytest.mark.asyncio
async def test_get_balance_throws_when_token_not_specified(account):
    modified_account = Account(
        address=account.address,
        client=FullNodeClient(node_url="custom.net/rpc"),
        key_pair=KeyPair(1, 2),
        chain=cast(StarknetChainId, 1),
    )
    with pytest.raises(
        ValueError,
        match="Argument token_address must be specified when using a custom network.",
    ):
        await modified_account.get_balance()


@pytest.mark.skipif(
    "--contract_dir=v1" in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.asyncio
async def test_balance_when_token_specified(account, erc20_contract):
    balance = await account.get_balance(erc20_contract.address)

    assert balance == 200


@pytest.mark.skipif(
    "--contract_dir=v1" in sys.argv,
    reason="Contract exists only in v2 directory",
)
@pytest.mark.asyncio
async def test_estimated_fee_greater_than_zero(account, erc20_contract):
    estimated_fee = (
        await erc20_contract.functions["balance_of"]
        .prepare_invoke_v3(
            account.address, resource_bounds=ResourceBoundsMapping.init_with_zeros()
        )
        .estimate_fee(block_hash="latest")
    )

    assert estimated_fee.overall_fee > 0
    assert estimated_fee.calculate_overall_fee() == estimated_fee.overall_fee


@pytest.mark.asyncio
async def test_estimate_fee_for_declare_transaction(
    account, map_compiled_contract_and_class_hash
):
    (compiled_contract, class_hash) = map_compiled_contract_and_class_hash
    declare_tx = await account.sign_declare_v3(
        compiled_contract=compiled_contract,
        compiled_class_hash=class_hash,
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )

    estimated_fee = await account.client.estimate_fee(tx=declare_tx)

    assert isinstance(estimated_fee.overall_fee, int)
    assert estimated_fee.overall_fee > 0
    assert estimated_fee.calculate_overall_fee() == estimated_fee.overall_fee


@pytest.mark.asyncio
async def test_account_estimate_fee_for_declare_transaction(
    account, map_compiled_contract_and_class_hash
):
    (compiled_contract, class_hash) = map_compiled_contract_and_class_hash
    declare_tx = await account.sign_declare_v3(
        compiled_contract=compiled_contract,
        compiled_class_hash=class_hash,
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )

    estimated_fee = await account.estimate_fee(tx=declare_tx)

    assert estimated_fee.unit == PriceUnit.FRI
    assert isinstance(estimated_fee.overall_fee, int)
    assert estimated_fee.overall_fee > 0
    assert estimated_fee.calculate_overall_fee() == estimated_fee.overall_fee


@pytest.mark.asyncio
async def test_account_estimate_fee_for_transactions(account, map_contract):
    invoke_tx_1 = await account.sign_invoke_v3(
        calls=Call(map_contract.address, get_selector_from_name("put"), [3, 4]),
        resource_bounds=MAX_RESOURCE_BOUNDS,
        nonce=(await account.get_nonce()),
    )

    invoke_tx_2 = await account.sign_invoke_v3(
        calls=Call(map_contract.address, get_selector_from_name("put"), [5, 1]),
        resource_bounds=MAX_RESOURCE_BOUNDS,
        nonce=(await account.get_nonce() + 1),
    )

    estimated_fee = await account.estimate_fee(tx=[invoke_tx_1, invoke_tx_2])

    assert len(estimated_fee) == 2
    assert isinstance(estimated_fee[0], EstimatedFee)
    assert isinstance(estimated_fee[1], EstimatedFee)
    assert estimated_fee[0].unit == PriceUnit.FRI
    assert estimated_fee[1].unit == PriceUnit.FRI
    assert isinstance(estimated_fee[0].overall_fee, int)
    assert estimated_fee[0].overall_fee > 0
    assert estimated_fee[0].calculate_overall_fee() == estimated_fee[0].overall_fee


@pytest.mark.asyncio
@pytest.mark.parametrize("key, val", [(20, 20), (30, 30)])
async def test_sending_multicall(account, map_contract, key, val):
    calls = [
        map_contract.functions["put"].prepare_invoke_v3(key=10, value=10),
        map_contract.functions["put"].prepare_invoke_v3(key=key, value=val),
    ]

    res = await account.execute_v3(calls=calls, resource_bounds=MAX_RESOURCE_BOUNDS)
    await account.client.wait_for_tx(res.transaction_hash)
    (value,) = await map_contract.functions["get"].call(key=key)

    assert value == val


@pytest.mark.asyncio
async def test_rejection_reason_in_transaction_receipt(map_contract):
    with pytest.raises(
        ClientError,
        match="The transaction's resources don't cover validation or the minimal transaction fee",
    ):
        resource_bounds = ResourceBoundsMapping(
            l1_gas=ResourceBounds(max_amount=1, max_price_per_unit=1),
            l2_gas=ResourceBounds(max_amount=1, max_price_per_unit=1),
            l1_data_gas=ResourceBounds(max_amount=1, max_price_per_unit=1),
        )
        await map_contract.functions["put"].invoke_v3(
            key=10, value=20, resource_bounds=resource_bounds
        )


def test_sign_and_verify_offchain_message_fail(account, typed_data):
    signature = account.sign_message(typed_data)
    signature = [signature[0] + 1, signature[1]]
    result = account.verify_message(typed_data, signature)

    assert result is False


def test_sign_and_verify_offchain_message(account, typed_data):
    signature = account.sign_message(typed_data)
    result = account.verify_message(typed_data, signature)

    assert result is True


@pytest.mark.asyncio
async def test_get_class_hash_at(account, map_contract):
    class_hash = await account.client.get_class_hash_at(
        map_contract.address, block_hash="latest"
    )

    assert class_hash != 0


@pytest.mark.asyncio()
async def test_get_nonce(account, map_contract):
    nonce = await account.get_nonce()
    address = map_contract.address
    block = await account.client.get_block(block_number="latest")

    tx = await account.execute_v3(
        Call(
            to_addr=address, selector=get_selector_from_name("put"), calldata=[10, 20]
        ),
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )
    await account.client.wait_for_tx(tx.transaction_hash)

    new_nonce = await account.get_nonce()
    new_nonce_latest_block = await account.get_nonce(block_number="latest")

    old_nonce = await account.get_nonce(block_number=block.block_number)

    assert isinstance(nonce, int) and isinstance(new_nonce, int)
    assert new_nonce == nonce + 1

    assert old_nonce == nonce
    assert new_nonce_latest_block == new_nonce


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "calls", [[Call(10, 20, [30])], [Call(10, 20, [30]), Call(40, 50, [60])]]
)
async def test_sign_invoke_v3(account, calls):
    signed_tx = await account.sign_invoke_v3(calls, resource_bounds=MAX_RESOURCE_BOUNDS)

    assert isinstance(signed_tx, InvokeV3)
    assert isinstance(signed_tx.signature, list)
    assert len(signed_tx.signature) == 2
    assert signed_tx.resource_bounds == MAX_RESOURCE_BOUNDS
    assert signed_tx.version == 3


@pytest.mark.asyncio
async def test_sign_invoke_v3_auto_estimate(account, map_contract):
    signed_tx = await account.sign_invoke_v3(
        Call(map_contract.address, get_selector_from_name("put"), [3, 4]),
        auto_estimate=True,
    )

    assert isinstance(signed_tx, InvokeV3)
    assert signed_tx.version == 3

    assert isinstance(signed_tx.signature, list)
    assert len(signed_tx.signature) == 2

    assert isinstance(signed_tx.resource_bounds, ResourceBoundsMapping)
    assert signed_tx.resource_bounds.l1_gas.max_amount >= 0
    assert signed_tx.resource_bounds.l1_gas.max_price_per_unit > 0
    assert signed_tx.resource_bounds.l2_gas.max_amount >= 0
    assert signed_tx.resource_bounds.l2_gas.max_price_per_unit > 0
    assert signed_tx.resource_bounds.l1_data_gas.max_amount >= 0
    assert signed_tx.resource_bounds.l1_data_gas.max_price_per_unit > 0


@pytest.mark.asyncio
async def test_sign_declare_v3(
    account, sierra_minimal_compiled_contract_and_class_hash
):
    (
        compiled_contract,
        compiled_class_hash,
    ) = sierra_minimal_compiled_contract_and_class_hash
    signed_tx = await account.sign_declare_v3(
        compiled_contract,
        compiled_class_hash,
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )

    assert isinstance(signed_tx, DeclareV3)
    assert signed_tx.version == 3
    assert isinstance(signed_tx.signature, list)
    assert len(signed_tx.signature) == 2
    assert signed_tx.nonce is not None
    assert signed_tx.resource_bounds == MAX_RESOURCE_BOUNDS
    assert signed_tx.version == 3


@pytest.mark.asyncio
async def test_sign_declare_v3_auto_estimate(
    account, sierra_minimal_compiled_contract_and_class_hash
):
    (
        compiled_contract,
        compiled_class_hash,
    ) = sierra_minimal_compiled_contract_and_class_hash

    signed_tx = await account.sign_declare_v3(
        compiled_contract, compiled_class_hash, auto_estimate=True
    )

    assert isinstance(signed_tx, DeclareV3)
    assert signed_tx.version == 3

    assert isinstance(signed_tx.signature, list)
    assert len(signed_tx.signature) == 2

    assert isinstance(signed_tx.resource_bounds, ResourceBoundsMapping)
    assert signed_tx.resource_bounds.l1_gas.max_amount >= 0
    assert signed_tx.resource_bounds.l1_gas.max_price_per_unit > 0
    assert signed_tx.resource_bounds.l2_gas.max_amount >= 0
    assert signed_tx.resource_bounds.l2_gas.max_price_per_unit > 0
    assert signed_tx.resource_bounds.l1_data_gas.max_amount >= 0
    assert signed_tx.resource_bounds.l1_data_gas.max_price_per_unit > 0


@pytest.mark.asyncio
async def test_sign_deploy_account_v3(account):
    class_hash = 0x1234
    salt = 0x123
    calldata = [1, 2, 3]
    signed_tx = await account.sign_deploy_account_v3(
        class_hash,
        salt,
        resource_bounds=MAX_RESOURCE_BOUNDS,
        constructor_calldata=calldata,
    )

    assert isinstance(signed_tx, DeployAccountV3)
    assert signed_tx.version == 3

    assert isinstance(signed_tx.signature, list)
    assert len(signed_tx.signature) == 2

    assert signed_tx.resource_bounds == MAX_RESOURCE_BOUNDS
    assert signed_tx.class_hash == class_hash
    assert signed_tx.contract_address_salt == salt
    assert signed_tx.constructor_calldata == calldata


@pytest.mark.asyncio
async def test_sign_deploy_account_v3_auto_estimate(
    account, account_with_validate_deploy_class_hash
):
    class_hash = account_with_validate_deploy_class_hash
    salt = 0x123
    calldata = [account.signer.public_key]
    signed_tx = await account.sign_deploy_account_v3(
        class_hash, salt, constructor_calldata=calldata, auto_estimate=True
    )

    assert isinstance(signed_tx, DeployAccountV3)
    assert signed_tx.version == 3

    assert isinstance(signed_tx.signature, list)
    assert len(signed_tx.signature) == 2

    assert isinstance(signed_tx.resource_bounds, ResourceBoundsMapping)
    assert signed_tx.resource_bounds.l1_gas.max_amount >= 0
    assert signed_tx.resource_bounds.l1_gas.max_price_per_unit > 0
    assert signed_tx.resource_bounds.l2_gas.max_amount >= 0
    assert signed_tx.resource_bounds.l2_gas.max_price_per_unit > 0
    assert signed_tx.resource_bounds.l1_data_gas.max_amount >= 0
    assert signed_tx.resource_bounds.l1_data_gas.max_price_per_unit > 0


@pytest.mark.asyncio
async def test_deploy_account_v3(client, deploy_account_details_factory):
    address, key_pair, salt, class_hash = await deploy_account_details_factory.get()

    deploy_result = await Account.deploy_account_v3(
        address=address,
        class_hash=class_hash,
        salt=salt,
        key_pair=key_pair,
        client=client,
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )
    await deploy_result.wait_for_acceptance()

    account = deploy_result.account

    assert isinstance(account, BaseAccount)
    assert account.address == address

    transaction = await client.get_transaction(tx_hash=deploy_result.hash)
    assert isinstance(transaction, DeployAccountTransactionV3)
    assert transaction.constructor_calldata == [key_pair.public_key]


@pytest.mark.asyncio
async def test_deploy_account_raises_on_incorrect_address(
    client, deploy_account_details_factory
):
    address, key_pair, salt, class_hash = await deploy_account_details_factory.get()

    with pytest.raises(
        ValueError,
        match=f"Provided address {hex(0x111)} is different than computed address {hex(address)}",
    ):
        await Account.deploy_account_v3(
            address=0x111,
            class_hash=class_hash,
            salt=salt,
            key_pair=key_pair,
            client=client,
            resource_bounds=MAX_RESOURCE_BOUNDS,
        )


@pytest.mark.asyncio
@pytest.mark.skip("TODO(#1560)")
async def test_deploy_account_raises_on_no_enough_funds(
    deploy_account_details_factory, client
):
    address, key_pair, salt, class_hash = await deploy_account_details_factory.get()

    with patch(
        f"{FullNodeClient.__module__}.FullNodeClient.call_contract", AsyncMock()
    ) as mocked_balance:
        mocked_balance.return_value = (0, 0)

        with pytest.raises(
            ValueError,
            match="Not enough tokens at the specified address to cover deployment costs",
        ):
            await Account.deploy_account_v3(
                address=address,
                class_hash=class_hash,
                salt=salt,
                key_pair=key_pair,
                client=client,
                resource_bounds=MAX_RESOURCE_BOUNDS,
            )


@pytest.mark.asyncio
async def test_deploy_account_passes_on_enough_funds(
    deploy_account_details_factory, client
):
    address, key_pair, salt, class_hash = await deploy_account_details_factory.get()

    with patch(
        f"{FullNodeClient.__module__}.FullNodeClient.call_contract", AsyncMock()
    ) as mocked_balance, patch(
        f"{FullNodeClient.__module__}.FullNodeClient.deploy_account", AsyncMock()
    ) as mocked_deploy:
        mocked_balance.return_value = (0, 100)
        mocked_deploy.return_value = DeployAccountTransactionResponse(
            transaction_hash=0x1
        )

        await Account.deploy_account_v3(
            address=address,
            class_hash=class_hash,
            salt=salt,
            key_pair=key_pair,
            client=client,
            resource_bounds=MAX_RESOURCE_BOUNDS,
        )


# TODO (#1056): change this test to braavos account
@pytest.mark.skip(
    reason="'__validate_execute__' doesn't allow any other calldata than in the constructor"
)
@pytest.mark.asyncio
async def test_deploy_account_uses_custom_calldata(
    client, deploy_account_details_factory, fee_contract
):
    _, key_pair, salt, class_hash = await deploy_account_details_factory.get()
    calldata = [1, 2, 3, 4]
    address = compute_address(
        salt=salt,
        class_hash=class_hash,
        constructor_calldata=calldata,
        deployer_address=0,
    )

    res = await fee_contract.functions["transfer"].invoke(
        recipient=address, amount=int(1e16), resource_bounds=MAX_RESOURCE_BOUNDS
    )
    await res.wait_for_acceptance()

    deploy_result = await Account.deploy_account_v3(
        address=address,
        class_hash=class_hash,
        salt=salt,
        key_pair=key_pair,
        client=client,
        constructor_calldata=calldata,
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )

    tx = await client.get_transaction(deploy_result.hash)
    assert isinstance(tx, DeployAccountTransactionV3)
    assert tx.constructor_calldata == calldata


@pytest.mark.asyncio
async def test_sign_invoke_v3_for_fee_estimation(account, map_contract):
    call = map_contract.functions["put"].prepare_invoke_v3(key=1, value=2)
    transaction = await account.sign_invoke_v3(
        calls=call, resource_bounds=MAX_RESOURCE_BOUNDS
    )

    estimate_fee_transaction = await account.sign_for_fee_estimate(transaction)
    assert estimate_fee_transaction.version == transaction.version + 2**128

    estimation = await account.client.estimate_fee(estimate_fee_transaction)
    assert isinstance(estimation, EstimatedFee)
    assert estimation.unit == PriceUnit.FRI
    assert estimation.overall_fee > 0


@pytest.mark.asyncio
async def test_sign_transaction_custom_nonce(account, hello_starknet_class_hash):
    deployment = Deployer().create_contract_deployment(hello_starknet_class_hash)
    deploy_tx = await account.sign_invoke_v3(
        deployment.call, resource_bounds=MAX_RESOURCE_BOUNDS
    )

    new_balance = 30
    invoke_tx = await account.sign_invoke_v3(
        Call(
            deployment.address,
            get_selector_from_name("increase_balance"),
            [new_balance],
        ),
        nonce=deploy_tx.nonce + 1,
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )

    deploy_res = await account.client.send_transaction(deploy_tx)
    invoke_res = await account.client.send_transaction(invoke_tx)

    await account.client.wait_for_tx(deploy_res.transaction_hash)
    await account.client.wait_for_tx(invoke_res.transaction_hash)

    result = await account.client.call_contract(
        Call(deployment.address, get_selector_from_name("get_balance"), [])
    )

    assert invoke_tx.nonce == deploy_tx.nonce + 1
    assert result == [new_balance]


@pytest.mark.asyncio
async def test_argent_account_deploy(
    client,
    argent_account_v040_data: AccountPrerequisites,
):
    deploy_result = await Account.deploy_account_v3(
        address=argent_account_v040_data.address,
        class_hash=ARGENT_V040_CLASS_HASH,
        salt=argent_account_v040_data.salt,
        key_pair=argent_account_v040_data.key_pair,
        client=client,
        constructor_calldata=argent_account_v040_data.calldata,
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )

    await deploy_result.wait_for_acceptance()
    account = deploy_result.account

    assert isinstance(account, BaseAccount)
    assert await account.cairo_version == 1

    account_contract_class = await client.get_class_at(
        contract_address=account.address, block_number="latest"
    )

    assert isinstance(account_contract_class, SierraContractClass)


@pytest.mark.asyncio
async def test_argent_account_execute(
    deployed_balance_contract,
    argent_account_v040: BaseAccount,
):
    # verify that initial balance is 0
    get_balance_call = Call(
        to_addr=deployed_balance_contract.address,
        selector=get_selector_from_name("get_balance"),
        calldata=[],
    )
    get_balance = await argent_account_v040.client.call_contract(
        call=get_balance_call, block_number="latest"
    )

    assert get_balance[0] == 0

    value = 20
    increase_balance_by_20_call = Call(
        to_addr=deployed_balance_contract.address,
        selector=get_selector_from_name("increase_balance"),
        calldata=[value],
    )
    execute = await argent_account_v040.execute_v3(
        calls=increase_balance_by_20_call, resource_bounds=MAX_RESOURCE_BOUNDS
    )
    await argent_account_v040.client.wait_for_tx(tx_hash=execute.transaction_hash)
    receipt = await argent_account_v040.client.get_transaction_receipt(
        tx_hash=execute.transaction_hash
    )

    assert receipt.finality_status == TransactionFinalityStatus.ACCEPTED_ON_L2

    # verify that the previous call was executed
    get_balance_call = Call(
        to_addr=deployed_balance_contract.address,
        selector=get_selector_from_name("get_balance"),
        calldata=[],
    )
    get_balance = await argent_account_v040.client.call_contract(
        call=get_balance_call, block_number="latest"
    )

    assert get_balance[0] == value


@pytest.mark.asyncio
async def test_account_execute_v3(account, deployed_balance_contract):
    get_balance_call = Call(
        to_addr=deployed_balance_contract.address,
        selector=get_selector_from_name("get_balance"),
        calldata=[],
    )
    increase_balance_call = Call(
        to_addr=deployed_balance_contract.address,
        selector=get_selector_from_name("increase_balance"),
        calldata=[100],
    )

    (initial_balance,) = await account.client.call_contract(call=get_balance_call)

    execute_increase_balance = await account.execute_v3(
        calls=increase_balance_call, resource_bounds=MAX_RESOURCE_BOUNDS
    )
    receipt = await account.client.wait_for_tx(
        tx_hash=execute_increase_balance.transaction_hash
    )

    assert receipt.execution_status == TransactionExecutionStatus.SUCCEEDED

    tx_details = await account.client.get_transaction(
        tx_hash=execute_increase_balance.transaction_hash
    )
    assert isinstance(tx_details, InvokeTransactionV3)

    (balance_after_increase,) = await account.client.call_contract(
        call=get_balance_call
    )
    assert initial_balance + 100 == balance_after_increase


@pytest.mark.asyncio
async def test_invoke_v3_with_tip(account, hello_starknet_class_hash):
    deployment = Deployer().create_contract_deployment(hello_starknet_class_hash)

    invoke_tx = await account.execute_v3(
        Call(
            deployment.address,
            get_selector_from_name("increase_balance"),
            [20000],
        ),
        resource_bounds=MAX_RESOURCE_BOUNDS,
        tip=123456,
    )

    transaction = await account.client.get_transaction(
        tx_hash=invoke_tx.transaction_hash
    )

    assert isinstance(transaction, InvokeTransactionV3)
    assert transaction.tip == 123456


@pytest.mark.asyncio
async def test_invoke_v3_auto_estimate_tip(
    account, hello_starknet_class_hash, get_block_with_txs_path, block_with_tips_mock
):
    with patch(get_block_with_txs_path, AsyncMock()) as mocked_block_with_txs:
        mocked_block_with_txs.return_value = block_with_tips_mock

        deployment = Deployer().create_contract_deployment(hello_starknet_class_hash)

        invoke_tx = await account.execute_v3(
            Call(
                deployment.address,
                get_selector_from_name("increase_balance"),
                [20000],
            ),
            resource_bounds=MAX_RESOURCE_BOUNDS,
            auto_estimate_tip=True,
        )

    transaction = await account.client.get_transaction(
        tx_hash=invoke_tx.transaction_hash
    )

    assert isinstance(transaction, InvokeTransactionV3)
    assert transaction.tip == 2


@pytest.mark.asyncio
async def test_deploy_account_v3_with_tip(client, deploy_account_details_factory):
    address, key_pair, salt, class_hash = await deploy_account_details_factory.get()

    tip = 12345
    deploy_result = await Account.deploy_account_v3(
        address=address,
        class_hash=class_hash,
        salt=salt,
        key_pair=key_pair,
        client=client,
        resource_bounds=MAX_RESOURCE_BOUNDS,
        tip=tip,
    )
    await deploy_result.wait_for_acceptance()

    transaction = await client.get_transaction(tx_hash=deploy_result.hash)
    assert isinstance(transaction, DeployAccountTransactionV3)
    assert transaction.tip == tip


@pytest.mark.asyncio
async def test_deploy_account_v3_auto_estimate_tip(
    client,
    deploy_account_details_factory,
    get_block_with_txs_path,
    block_with_tips_mock,
):
    address, key_pair, salt, class_hash = await deploy_account_details_factory.get()

    with patch(get_block_with_txs_path, AsyncMock()) as mocked_block_with_txs:
        mocked_block_with_txs.return_value = block_with_tips_mock

        deploy_result = await Account.deploy_account_v3(
            address=address,
            class_hash=class_hash,
            salt=salt,
            key_pair=key_pair,
            client=client,
            resource_bounds=MAX_RESOURCE_BOUNDS,
            auto_estimate_tip=True,
        )
        await deploy_result.wait_for_acceptance()

    transaction = await client.get_transaction(tx_hash=deploy_result.hash)
    assert isinstance(transaction, DeployAccountTransactionV3)
    assert transaction.tip == 2


@pytest.mark.asyncio
async def test_declare_v3_with_tip(account):
    compiled_contract = load_contract("TestContract3")

    tip = 12345
    signed_tx = await account.sign_declare_v3(
        compiled_contract["sierra"],
        compute_casm_class_hash(create_casm_class(compiled_contract["casm"])),
        resource_bounds=MAX_RESOURCE_BOUNDS,
        tip=tip,
    )

    result = await account.client.declare(signed_tx)
    transaction = await account.client.get_transaction(tx_hash=result.transaction_hash)
    assert isinstance(transaction, DeclareTransactionV3)
    assert transaction.tip == tip


@pytest.mark.asyncio
async def test_declare_v3_auto_estimate_tip(
    account,
    get_block_with_txs_path,
    block_with_tips_mock,
):
    compiled_contract = load_contract("TestContract4")

    with patch(get_block_with_txs_path, AsyncMock()) as mocked_block_with_txs:
        mocked_block_with_txs.return_value = block_with_tips_mock
        signed_tx = await account.sign_declare_v3(
            compiled_contract["sierra"],
            compute_casm_class_hash(create_casm_class(compiled_contract["casm"])),
            resource_bounds=MAX_RESOURCE_BOUNDS,
            auto_estimate_tip=True,
        )

    result = await account.client.declare(signed_tx)
    transaction = await account.client.get_transaction(tx_hash=result.transaction_hash)
    assert isinstance(transaction, DeclareTransactionV3)
    assert transaction.tip == 2



import json
import re
from enum import Enum
from typing import AsyncGenerator, List, Tuple, TypedDict, Union, cast

from starknet_py.abi.v0.shape import AbiDictList
from starknet_py.constants import (
    RPC_CLASS_HASH_NOT_FOUND_ERROR,
    RPC_CONTRACT_ERROR,
    RPC_CONTRACT_NOT_FOUND_ERROR,
    RPC_INVALID_MESSAGE_SELECTOR_ERROR,
)
from starknet_py.net.client import Client
from starknet_py.net.client_errors import ClientError, ContractNotFoundError
from starknet_py.net.client_models import DeprecatedContractClass, SierraContractClass
from starknet_py.net.models import Address
from starknet_py.proxy.proxy_check import (
    ArgentProxyCheck,
    OpenZeppelinProxyCheck,
    ProxyCheck,
)


class ProxyConfig(TypedDict, total=False):
    """
    Proxy resolving configuration.
    """

    proxy_checks: List[ProxyCheck]
    """
    List of classes implementing :class:`~starknet_py.proxy.proxy_check.ProxyCheck` ABC,
    that will be used for checking if contract at the address is a proxy contract.
    """


def prepare_proxy_config(proxy_config: ProxyConfig) -> ProxyConfig:
    if "proxy_checks" in proxy_config:
        return proxy_config

    proxy_checks = [
        OpenZeppelinProxyCheck(),
        ArgentProxyCheck(),
    ]
    return {"proxy_checks": proxy_checks}


class ImplementationType(Enum):
    """
    Enum representing transaction types
    """

    CLASS_HASH = "class_hash"
    ADDRESS = "address"


class ContractAbiResolver:
    """
    Class for resolving abi of a contract
    """

    def __init__(
        self,
        address: Address,
        client: Client,
        proxy_config: ProxyConfig,
    ):
        """
        :param address: Contract's address
        :param client: Client used for resolving abi
        :param proxy_config: Proxy config for resolving proxy
        """
        self.address = address
        self.client = client
        self.proxy_config = proxy_config

    async def resolve(self) -> Tuple[AbiDictList, int]:
        """
        Returns abi and cairo version of either direct contract
        or contract proxied by direct contract depending on proxy_config.

        :raises ContractNotFoundError: when contract could not be found at address
        :raises ProxyResolutionError: when given ProxyChecks were not sufficient to resolve proxy
        :raises AbiNotFoundError: when abi is not present in contract class at address
        """
        if len(self.proxy_config) == 0:
            return await self.get_abi_for_address()
        return await self.resolve_abi()

    async def get_abi_for_address(self) -> Tuple[AbiDictList, int]:
        """
        Returns abi and cairo version of a contract directly from address.

        :raises ContractNotFoundError: when contract could not be found at address
        :raises AbiNotFoundError: when abi is not present in contract class at address
        """
        contract_class = await _get_class_at(address=self.address, client=self.client)

        if contract_class.abi is None:
            raise AbiNotFoundError()

        return self.get_abi_from_contract_class(
            contract_class
        ), self._get_cairo_version(contract_class)

    async def resolve_abi(self) -> Tuple[AbiDictList, int]:
        """
        Returns abi and cairo version of a contract that is being proxied by contract at address.

        :raises ContractNotFoundError: when contract could not be found at address
        :raises ProxyResolutionError: when given ProxyChecks were not sufficient to resolve proxy
        :raises AbiNotFoundError: when abi is not present in proxied contract class at address
        """
        implementation_generator = self._get_implementation_from_proxy()

        # implementation is either a class_hash or address
        async for implementation, implementation_type in implementation_generator:
            try:
                if implementation_type == ImplementationType.CLASS_HASH:
                    contract_class = await self.client.get_class_by_hash(implementation)
                else:
                    contract_class = await _get_class_at(
                        address=implementation, client=self.client
                    )

                if contract_class.abi is None:
                    # Some contract_class has been found, but it does not have abi
                    raise AbiNotFoundError()

                return self.get_abi_from_contract_class(
                    contract_class
                ), self._get_cairo_version(contract_class)
            except ClientError as err:
                if not (
                    "is not declared" in err.message
                    or err.code == RPC_CLASS_HASH_NOT_FOUND_ERROR
                    or isinstance(err, ContractNotFoundError)
                ):
                    raise err

        raise ProxyResolutionError(self.proxy_config.get("proxy_checks", []))

    @staticmethod
    def _get_cairo_version(
        contract_class: Union[DeprecatedContractClass, SierraContractClass]
    ) -> int:
        return 1 if isinstance(contract_class, SierraContractClass) else 0

    @staticmethod
    def get_abi_from_contract_class(
        contract_class: Union[DeprecatedContractClass, SierraContractClass]
    ) -> AbiDictList:
        return (
            cast(AbiDictList, contract_class.abi)
            if isinstance(contract_class, DeprecatedContractClass)
            else json.loads(cast(str, contract_class.abi))
        )

    async def _get_implementation_from_proxy(
        self,
    ) -> AsyncGenerator[Tuple[int, ImplementationType], None]:
        proxy_checks = self.proxy_config.get("proxy_checks", [])
        for proxy_check in proxy_checks:
            try:
                implementation = await proxy_check.implementation_hash(
                    address=self.address, client=self.client
                )
                if implementation is not None:
                    yield implementation, ImplementationType.CLASS_HASH

                implementation = await proxy_check.implementation_address(
                    address=self.address, client=self.client
                )
                if implementation is not None:
                    yield implementation, ImplementationType.ADDRESS
            except ClientError as err:
                err_msg = (
                    r"(Entry point ((0x[0-9a-f]+)|(EntryPointSelector\(StarkFelt\(\"0x[0-9a-f]+)\"\)\))"
                    r" not found in contract)|(is not declared)|(is not deployed)"
                )
                if not (
                    re.search(err_msg, err.message, re.IGNORECASE)
                    or err.code
                    in [
                        RPC_CLASS_HASH_NOT_FOUND_ERROR,
                        RPC_CONTRACT_NOT_FOUND_ERROR,
                        RPC_CONTRACT_ERROR,
                        RPC_INVALID_MESSAGE_SELECTOR_ERROR,  # removed in RPC v0.3.0, backwards compatibility for nodes
                    ]
                ):
                    raise err


class AbiNotFoundError(Exception):
    """
    Error while resolving contract abi.
    """


class ProxyResolutionError(Exception):
    """
    Error while resolving proxy using ProxyChecks.
    """

    DOCS = "https://starknetpy.readthedocs.io/en/latest/guide/resolving_proxy_contracts.html#proxychecks"

    def __init__(
        self,
        proxy_checks: List[ProxyCheck],
    ):
        proxy_check_names = tuple(
            proxy_check.__class__.__name__ for proxy_check in proxy_checks
        )
        proxy_checks_str = " " + str(proxy_check_names) if proxy_check_names else ""
        self.message = f"""Couldn't resolve proxy using given ProxyChecks{proxy_checks_str}.
        See {self.DOCS} for a guide on writing own ProxyChecks."""
        super().__init__(self.message)


async def _get_class_at(
    address: Address, client: Client
) -> Union[DeprecatedContractClass, SierraContractClass]:
    try:
        contract_class_hash = await client.get_class_hash_at(contract_address=address)
        contract_class = await client.get_class_by_hash(class_hash=contract_class_hash)
    except ClientError as err:
        if (
            "is not deployed" in err.message
            or err.code == RPC_CLASS_HASH_NOT_FOUND_ERROR
            or err.code == RPC_CONTRACT_NOT_FOUND_ERROR
        ):
            raise ContractNotFoundError(address=address) from err
        raise err

    return contract_class



from abc import ABC, abstractmethod
from typing import List, Optional, Union

from starknet_py.constants import ANY_CALLER, OutsideExecutionInterfaceID
from starknet_py.net.client import Client
from starknet_py.net.client_models import (
    Call,
    Calls,
    EstimatedFee,
    Hash,
    OutsideExecutionTimeBounds,
    ResourceBoundsMapping,
    SentTransactionResponse,
    Tag,
)
from starknet_py.net.models import AddressRepresentation
from starknet_py.net.models.transaction import (
    AccountTransaction,
    DeclareV3,
    DeployAccountV3,
    InvokeV3,
    TypeAccountTransaction,
)
from starknet_py.net.models.typed_data import TypedDataDict


class OutsideExecutionSupportBaseMixin(ABC):

    @abstractmethod
    async def get_outside_execution_nonce(self) -> int:
        """
        Generate special valid nonce for outside execution calls.
        """

    @abstractmethod
    async def supports_interface(
        self, interface_id: OutsideExecutionInterfaceID
    ) -> bool:
        """
        Check if the account supports the given outside execution interface. Part of ISRC5 standard.
        """

    @abstractmethod
    async def sign_outside_execution_call(
        self,
        calls: Calls,
        execution_time_bounds: OutsideExecutionTimeBounds,
        *,
        caller: AddressRepresentation = ANY_CALLER,
        nonce: Optional[int] = None,
        interface_version: Optional[OutsideExecutionInterfaceID] = None,
    ) -> Call:
        """
        Creates a call for an outcide execution (SNIP-9 specification).

        :param calls: Single call or list of calls to be executed by outside caller.
        :param execution_time_bounds: Execution time bounds for the call.
        :param caller: Address of the caller. IMPORTANT! By default it is ANY_CALLER.
        :param nonce: Nonce for the transaction. Is populated automatically if not provided.
        :param interface_version: Outside execution interface version. Method will check which version account
            supports and use the highest one and populate the value.
        """


class BaseAccount(OutsideExecutionSupportBaseMixin, ABC):
    """
    Base class for all account implementations.

    Signs, prepares and executes transactions.
    """

    @property
    @abstractmethod
    def address(self) -> int:
        """
        Get the address of the account
        """

    @property
    @abstractmethod
    async def cairo_version(self) -> int:
        """
        Get Cairo version of the account.
        """

    @property
    @abstractmethod
    def client(self) -> Client:
        """
        Get the Client used by the Account.
        """

    @abstractmethod
    async def estimate_fee(
        self,
        tx: Union[AccountTransaction, List[AccountTransaction]],
        skip_validate: bool = False,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> Union[EstimatedFee, List[EstimatedFee]]:
        """
        Estimates the resources required by a given sequence of transactions when applied on a given state.
        If one of the transactions reverts or fails due to any reason (e.g. validation failure or an internal error),
        a TRANSACTION_EXECUTION_ERROR is returned.
        For v0-2 transactions the estimate is given in Wei, and for v3 transactions it is given in Fri.

        :param tx: Transaction or list of transactions to estimate
        :param skip_validate: Flag checking whether the validation part of the transaction should be executed
        :param block_hash: Block hash or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :param block_number: Block number or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :return: Estimated fee or list of estimated fees for each transaction
        """

    @abstractmethod
    async def get_nonce(
        self,
        *,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> int:
        """
        Get the current nonce of the account.

        :param block_hash: Block's hash or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :param block_number: Block's number or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :return: nonce of the account.
        """

    @abstractmethod
    async def get_balance(
        self,
        token_address: Optional[AddressRepresentation] = None,
        *,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> int:
        """
        Checks account's balance of the specified token.
        By default, it uses the L2 ETH address for mainnet and sepolia networks.

        :param token_address: Address of the ERC20 contract.
        :param block_hash: Block's hash or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :param block_number: Block's number or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :return: Token balance.
        """

    @abstractmethod
    async def sign_for_fee_estimate(
        self, transaction: TypeAccountTransaction
    ) -> TypeAccountTransaction:
        """
        Sign a transaction for a purpose of only fee estimation.
        Should use a transaction version that is not executable on Starknet,
        calculated like ``transaction.version + 2 ** 128``.

        :param transaction: Transaction to be signed.
        :return: A signed Transaction that can only be used for fee estimation and cannot be executed.
        """

    @abstractmethod
    async def sign_invoke_v3(
        self,
        calls: Calls,
        *,
        nonce: Optional[int] = None,
        resource_bounds: Optional[ResourceBoundsMapping] = None,
        auto_estimate: bool = False,
        tip: Optional[int] = None,
        auto_estimate_tip: bool = False,
    ) -> InvokeV3:
        # pylint: disable=too-many-arguments
        """
        Takes calls and creates signed Invoke.

        :param calls: Single call or list of calls.
        :param nonce: Nonce of the transaction.
        :param resource_bounds: Resource limits (L1 and L2) that can be used in this transaction.
        :param auto_estimate: Use automatic fee estimation, not recommend as it may lead to high costs.
        :param tip: The tip amount to be added to the transaction fee.
        :param auto_estimate_tip: Use automatic tip estimation. Using this option may lead to higher costs.
        :return: Invoke created from the calls.
        """

    @abstractmethod
    async def sign_declare_v3(
        self,
        compiled_contract: str,
        compiled_class_hash: int,
        *,
        nonce: Optional[int] = None,
        resource_bounds: Optional[ResourceBoundsMapping] = None,
        auto_estimate: bool = False,
        tip: Optional[int] = None,
        auto_estimate_tip: bool = False,
    ) -> DeclareV3:
        # pylint: disable=too-many-arguments
        """
        Create and sign declare transaction version 3 using sierra contract.

        :param compiled_contract: string containing a compiled Starknet contract.
            Supports new contracts (compiled to sierra).
        :param compiled_class_hash: a class hash of the sierra compiled contract used in the declare transaction.
            Computed from casm compiled contract.
        :param nonce: Nonce of the transaction.
        :param resource_bounds: Resource limits (L1 and L2) that can be used in this transaction.
        :param auto_estimate: Use automatic fee estimation, not recommend as it may lead to high costs.
        :param tip: The tip amount to be added to the transaction fee.
        :param auto_estimate_tip: Use automatic tip estimation. Using this option may lead to higher costs.
        :return: Signed DeclareV3 transaction.
        """

    @abstractmethod
    async def sign_deploy_account_v3(
        self,
        class_hash: int,
        contract_address_salt: int,
        *,
        constructor_calldata: Optional[List[int]] = None,
        nonce: int = 0,
        resource_bounds: Optional[ResourceBoundsMapping] = None,
        auto_estimate: bool = False,
        tip: Optional[int] = None,
        auto_estimate_tip: bool = False,
    ) -> DeployAccountV3:
        # pylint: disable=too-many-arguments
        """
        Create and sign deploy account transaction version 3.

        :param class_hash: Class hash of the contract class to be deployed.
        :param contract_address_salt: A salt used to calculate deployed contract address.
        :param constructor_calldata: Calldata to be ed to contract constructor
            and used to calculate deployed contract address.
        :param nonce: Nonce of the transaction.
        :param resource_bounds: Resource limits (L1 and L2) that can be used in this transaction.
            Enough tokens must be prefunded before sending the transaction for it to succeed.
        :param auto_estimate: Use automatic fee estimation, not recommend as it may lead to high costs.
        :param tip: The tip amount to be added to the transaction fee.
        :param auto_estimate_tip: Use automatic tip estimation. Using this option may lead to higher costs.
        :return: Signed DeployAccountV3 transaction.
        """

    @abstractmethod
    async def execute_v3(
        self,
        calls: Calls,
        *,
        resource_bounds: Optional[ResourceBoundsMapping] = None,
        nonce: Optional[int] = None,
        auto_estimate: bool = False,
        tip: Optional[int] = None,
        auto_estimate_tip: bool = False,
    ) -> SentTransactionResponse:
        # pylint: disable=too-many-arguments
        """
        Takes calls and executes transaction.

        :param calls: Single call or list of calls.
        :param resource_bounds: Resource limits (L1 and L2) that can be used in this transaction.
        :param nonce: Nonce of the transaction.
        :param auto_estimate: Use automatic fee estimation, not recommend as it may lead to high costs.
        :param tip: The tip amount to be added to the transaction fee.
        :param auto_estimate_tip: Use automatic tip estimation. Using this option may lead to higher costs.
        :return: SentTransactionResponse.
        """

    @abstractmethod
    def sign_message(self, typed_data: TypedDataDict) -> List[int]:
        """
        Sign an TypedData TypedDict for off-chain usage with the Starknet private key and return the signature.
        This adds a message prefix, so it can't be interchanged with transactions.
        Both v0 and v1 domain revision versions are supported.

        :param typed_data: TypedData TypedDict to be signed.
        :return: The signature of the TypedData TypedDict.
        """

    @abstractmethod
    def verify_message(self, typed_data: TypedDataDict, signature: List[int]) -> bool:
        """
        Verify a signature of a TypedData dict on Starknet.
        Both v0 and v1 domain revision versions are supported.

        :param typed_data: TypedData TypedDict to be verified.
        :param signature: signature of the TypedData TypedDict.
        :return: true if the signature is valid, false otherwise.
        """



from .address import Address, AddressRepresentation, parse_address
from .chains import StarknetChainId, chain_from_network
from .transaction import (
    AccountTransaction,
    DeclareV1,
    DeclareV2,
    DeclareV3,
    DeployAccountV1,
    DeployAccountV3,
    InvokeV1,
    InvokeV3,
    Transaction,
)



import pytest

from starknet_py.net.account.account import Account
from starknet_py.net.models.transaction import DeclareV3, DeployAccountV3, InvokeV3
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS


@pytest.mark.asyncio
async def test_account_sign_without_execute(
    account, sierra_minimal_compiled_contract_and_class_hash
):
    # pylint: disable=import-outside-toplevel
    assert isinstance(account, Account)
    address = selector = class_hash = salt = 0x1
    calldata = []
    (
        compiled_contract,
        compiled_class_hash,
    ) = sierra_minimal_compiled_contract_and_class_hash
    resource_bounds = MAX_RESOURCE_BOUNDS

    # docs: start
    from starknet_py.net.client_models import Call

    # Create a signed Invoke transaction
    call = Call(to_addr=address, selector=selector, calldata=calldata)
    invoke_transaction = await account.sign_invoke_v3(
        call, resource_bounds=resource_bounds
    )

    # Create a signed Declare transaction
    declare_transaction = await account.sign_declare_v3(
        compiled_contract,
        compiled_class_hash=compiled_class_hash,
        auto_estimate=True,
    )

    # Create a signed DeployAccount transaction
    deploy_account_transaction = await account.sign_deploy_account_v3(
        class_hash=class_hash,
        contract_address_salt=salt,
        constructor_calldata=calldata,
        resource_bounds=resource_bounds,
    )
    # docs: end

    assert isinstance(invoke_transaction, InvokeV3)
    assert isinstance(declare_transaction, DeclareV3)
    assert isinstance(deploy_account_transaction, DeployAccountV3)



from typing import Any, Optional, Union

from marshmallow import ValidationError, fields, post_load

from starknet_py.net.client_models import (
    BinaryNode,
    ContractLeafData,
    ContractsProof,
    EdgeNode,
    GlobalRoots,
    NodeHashToNodeMappingItem,
    StorageProofResponse,
)
from starknet_py.net.schemas.common import Felt, NumberAsHex
from starknet_py.utils.schema import Schema


class BinaryNodeSchema(Schema):
    left = Felt(data_key="left", required=True)
    right = Felt(data_key="right", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> BinaryNode:
        return BinaryNode(**data)


class EdgeNodeSchema(Schema):
    path = NumberAsHex(data_key="path", required=True)
    length = fields.Integer(data_key="length", required=True)
    child = Felt(data_key="child", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> EdgeNode:
        return EdgeNode(**data)


class MerkleNodeSchema(Schema):
    left = Felt(data_key="left", required=False)
    right = Felt(data_key="right", required=False)
    path = NumberAsHex(data_key="path", required=False)
    length = fields.Integer(data_key="length", required=False)
    child = Felt(data_key="child", required=False)

    @post_load
    def make_dataclass(self, data, **kwargs) -> Union[BinaryNode, EdgeNode]:
        if "left" in data and "right" in data:
            return BinaryNode(**data)
        elif "path" in data and "length" in data and "child" in data:
            return EdgeNode(**data)
        raise ValidationError(f"Invalid data provided for MerkleNode: {data}.")


class NodeHashToNodeMappingItemSchema(Schema):
    node_hash = Felt(data_key="node_hash", required=True)
    node = fields.Nested(MerkleNodeSchema(), data_key="node", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> NodeHashToNodeMappingItem:
        return NodeHashToNodeMappingItem(**data)


class NodeHashToNodeMappingField(fields.Field):
    def _serialize(self, value: Any, attr: Optional[str], obj: Any, **kwargs):
        if value is None:
            return None
        if not isinstance(value, list):
            raise ValidationError(
                f"Invalid value provided for NodeHashToNodeMapping: {value}. Expected a list."
            )
        return [NodeHashToNodeMappingItemSchema().dump(item) for item in value]

    def _deserialize(self, value: Any, attr: Optional[str], data: Any, **kwargs):
        if value is None:
            return None
        if not isinstance(value, list):
            raise ValidationError(
                f"Invalid value provided for NodeHashToNodeMapping: {value}. Expected a list."
            )
        return [NodeHashToNodeMappingItemSchema().load(item) for item in value]


class ContractLeafDataSchema(Schema):
    nonce = Felt(data_key="nonce", required=True)
    class_hash = Felt(data_key="class_hash", required=True)
    storage_root = Felt(data_key="storage_root", required=False)

    @post_load
    def make_dataclass(self, data, **kwargs) -> ContractLeafData:
        return ContractLeafData(**data)


class GlobalRootsSchema(Schema):
    contracts_tree_root = Felt(data_key="contracts_tree_root", required=True)
    classes_tree_root = Felt(data_key="classes_tree_root", required=True)
    block_hash = Felt(data_key="block_hash", required=True)

    @post_load
    def make_dataclass(self, data, **kwargs) -> GlobalRoots:
        return GlobalRoots(**data)


class ContractsProofSchema(Schema):
    nodes = NodeHashToNodeMappingField(data_key="nodes", required=True)
    contract_leaves_data = fields.List(
        fields.Nested(ContractLeafDataSchema()),
        data_key="contract_leaves_data",
        required=True,
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> ContractsProof:
        return ContractsProof(**data)


class StorageProofResponseSchema(Schema):
    classes_proof = fields.List(
        fields.Nested(NodeHashToNodeMappingItemSchema()),
        data_key="classes_proof",
        required=True,
    )
    contracts_proof = fields.Nested(
        ContractsProofSchema(), data_key="contracts_proof", required=True
    )
    contracts_storage_proofs = fields.List(
        NodeHashToNodeMappingField(), data_key="contracts_storage_proofs", required=True
    )
    global_roots = fields.Nested(
        GlobalRootsSchema(), data_key="global_roots", required=True
    )

    @post_load
    def make_dataclass(self, data, **kwargs) -> StorageProofResponse:
        return StorageProofResponse(**data)



import dataclasses
from typing import Optional, Tuple

from lark import Token, Transformer, v_args

from starknet_py.cairo.deprecated_parse.cairo_types import (
    CairoType,
    ExprIdentifier,
    TypeCodeoffset,
    TypeFelt,
    TypeIdentifier,
    TypePointer,
    TypeStruct,
    TypeTuple,
)


@dataclasses.dataclass
class ParserContext:
    """
    Represents information that affects the parsing process.
    """

    # If True, treat type identifiers as resolved.
    resolved_types: bool = False


class ParserError(Exception):
    """
    Base exception for parsing process.
    """


@dataclasses.dataclass
class CommaSeparated:
    """
    Represents a list of comma separated values, such as expressions or types.
    """

    args: list
    has_trailing_comma: bool


class ParserTransformer(Transformer):
    """
    Transforms the lark tree into an AST based on the classes defined in cairo_types.py.
    """

    # pylint: disable=unused-argument, no-self-use

    def __init__(self):
        super().__init__()
        self.parser_context = ParserContext()

    def __default__(self, data: str, children, meta):
        raise TypeError(f"Unable to parse tree node of type {data}")

    def comma_separated(self, value) -> CommaSeparated:
        saw_comma = None
        args: list = []
        for v in value:
            if isinstance(v, Token) and v.type == "COMMA":
                if saw_comma is not False:
                    raise ParserError("Unexpected comma.")
                saw_comma = True
            else:
                if saw_comma is False:
                    raise ParserError("Expected a comma before this expression.")
                args.append(v)

                # Reset state.
                saw_comma = False

        if saw_comma is None:
            saw_comma = False

        return CommaSeparated(args=args, has_trailing_comma=saw_comma)

    # Types.

    @v_args(meta=True)
    def named_type(self, meta, value) -> TypeTuple.Item:
        name: Optional[str]
        if len(value) == 1:
            # Unnamed type.
            (typ,) = value
            name = None
            if isinstance(typ, ExprIdentifier):
                typ = self.type_struct([typ])
        elif len(value) == 2:
            # Named type.
            identifier, typ = value
            assert isinstance(identifier, ExprIdentifier)
            assert isinstance(typ, CairoType)
            if "." in identifier.name:
                raise ParserError("Unexpected . in name.")
            name = identifier.name
        else:
            raise NotImplementedError(f"Unexpected number of values. {value}")

        return TypeTuple.Item(name=name, typ=typ)

    @v_args(meta=True)
    def type_felt(self, meta, value):
        return TypeFelt()

    @v_args(meta=True)
    def type_codeoffset(self, meta, value):
        return TypeCodeoffset()

    def type_struct(self, value):
        assert len(value) == 1 and isinstance(value[0], ExprIdentifier)
        if self.parser_context.resolved_types:
            # If parser_context.resolved_types is True, assume that the type is a struct.
            return TypeStruct(scope=value[0].name)

        return TypeIdentifier(name=value[0].name)

    @v_args(meta=True)
    def type_pointer(self, meta, value):
        return TypePointer(pointee=value[0])

    @v_args(meta=True)
    def type_pointer2(self, meta, value):
        return TypePointer(pointee=TypePointer(pointee=value[0]))

    @v_args(meta=True)
    def type_tuple(self, meta, value: Tuple[CommaSeparated]):
        (lst,) = value
        return TypeTuple(members=lst.args, has_trailing_comma=lst.has_trailing_comma)

    @v_args(meta=True)
    def identifier(self, meta, value):
        return ExprIdentifier(name=".".join(x.value for x in value))

    @v_args(meta=True)
    def identifier_def(self, meta, value):
        return ExprIdentifier(name=value[0].value)



import datetime

import pytest

from starknet_py.constants import ANY_CALLER, OutsideExecutionInterfaceID
from starknet_py.hash.selector import get_selector_from_name
from starknet_py.net.account.account import BaseAccount
from starknet_py.net.client_models import Call, OutsideExecutionTimeBounds
from starknet_py.tests.e2e.fixtures.constants import MAX_RESOURCE_BOUNDS
from starknet_py.transaction_errors import TransactionRevertedError


@pytest.mark.asyncio
async def test_argent_account_outside_execution_compatibility(
    argent_account_v040: BaseAccount,
):
    for interface, supported in [
        (OutsideExecutionInterfaceID.V1, True),
        (OutsideExecutionInterfaceID.V2, True),
    ]:
        assert await argent_account_v040.supports_interface(interface) is supported


@pytest.mark.asyncio
async def test_account_outside_execution_any_caller(
    argent_account_v040: BaseAccount,
    account: BaseAccount,
    map_contract,
):
    put_call = Call(
        to_addr=map_contract.address,
        selector=get_selector_from_name("put"),
        calldata=[20, 20],
    )

    call = await argent_account_v040.sign_outside_execution_call(
        calls=[
            put_call,
        ],
        execution_time_bounds=OutsideExecutionTimeBounds(
            execute_after=datetime.datetime.now() - datetime.timedelta(hours=1),
            execute_before=datetime.datetime.now() + datetime.timedelta(hours=1),
        ),
        caller=ANY_CALLER,
    )

    tx = await account.execute_v3(
        calls=[call],
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )
    await account.client.wait_for_tx(tx.transaction_hash)


@pytest.mark.asyncio
async def test_account_outside_execution_for_invalid_caller(
    argent_account_v040: BaseAccount,
    account: BaseAccount,
    map_contract,
):
    random_address = 0x1234567890123456789012345678901234567890

    put_call = Call(
        to_addr=map_contract.address,
        selector=get_selector_from_name("put"),
        calldata=[20, 20],
    )

    call = await argent_account_v040.sign_outside_execution_call(
        calls=[
            put_call,
        ],
        execution_time_bounds=OutsideExecutionTimeBounds(
            execute_after=datetime.datetime.now() - datetime.timedelta(hours=1),
            execute_before=datetime.datetime.now() + datetime.timedelta(hours=1),
        ),
        caller=random_address,
    )

    tx = await account.execute_v3(
        calls=[call],
        resource_bounds=MAX_RESOURCE_BOUNDS,
    )

    with pytest.raises(TransactionRevertedError) as err:
        await argent_account_v040.client.wait_for_tx(tx.transaction_hash)

    assert "argent/invalid-caller" in err.value.message


@pytest.mark.asyncio
async def test_account_outside_execution_for_impossible_time_bounds(
    argent_account_v040: BaseAccount,
    account: BaseAccount,
    map_contract,
):
    put_call = Call(
        to_addr=map_contract.address,
        selector=get_selector_from_name("put"),
        calldata=[20, 20],
    )

    call = await argent_account_v040.sign_outside_execution_call(
        calls=[put_call],
        execution_time_bounds=OutsideExecutionTimeBounds(
            execute_after=datetime.datetime.now() - datetime.timedelta(days=10),
            execute_before=datetime.datetime.now() - datetime.timedelta(days=9),
        ),
        caller=ANY_CALLER,
    )

    tx = await account.execute_v3(calls=[call], resource_bounds=MAX_RESOURCE_BOUNDS)

    with pytest.raises(TransactionRevertedError) as err:
        await argent_account_v040.client.wait_for_tx(tx.transaction_hash)

    assert "argent/invalid-timestamp" in err.value.message


@pytest.mark.asyncio
async def test_account_outside_execution_by_itself_is_impossible(
    argent_account_v040: BaseAccount,
    map_contract,
):
    put_call = Call(
        to_addr=map_contract.address,
        selector=get_selector_from_name("put"),
        calldata=[20, 20],
    )

    call = await argent_account_v040.sign_outside_execution_call(
        calls=[put_call],
        execution_time_bounds=OutsideExecutionTimeBounds(
            execute_after=datetime.datetime.now() - datetime.timedelta(days=10),
            execute_before=datetime.datetime.now() - datetime.timedelta(days=9),
        ),
        caller=ANY_CALLER,
    )

    tx = await argent_account_v040.execute_v3(
        calls=[call], resource_bounds=MAX_RESOURCE_BOUNDS
    )

    with pytest.raises(TransactionRevertedError) as err:
        await argent_account_v040.client.wait_for_tx(tx.transaction_hash)

    assert "ReentrancyGuard: reentrant call" in err.value.message



from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from typing import List, Optional, Union

from starknet_py.net.client_errors import ClientError
from starknet_py.net.client_models import (
    BlockStateUpdate,
    BlockTransactionTrace,
    Call,
    ContractsStorageKeys,
    DeclareTransactionResponse,
    DeployAccountTransactionResponse,
    DeprecatedContractClass,
    EstimatedFee,
    Hash,
    LatestTag,
    MessageStatus,
    PreConfirmedBlockStateUpdate,
    PreConfirmedStarknetBlock,
    PreConfirmedStarknetBlockWithReceipts,
    PreConfirmedStarknetBlockWithTxHashes,
    SentTransactionResponse,
    SierraContractClass,
    StarknetBlock,
    StarknetBlockWithReceipts,
    StarknetBlockWithTxHashes,
    StorageProofResponse,
    Tag,
    Transaction,
    TransactionExecutionStatus,
    TransactionReceiptWithBlockInfo,
    TransactionStatus,
    TransactionStatusResponse,
)
from starknet_py.net.executable_models import CasmClass
from starknet_py.net.models.transaction import (
    AccountTransaction,
    DeclareV3,
    DeployAccountV3,
    InvokeV3,
)
from starknet_py.transaction_errors import (
    TransactionNotReceivedError,
    TransactionRevertedError,
)
from starknet_py.utils.sync import add_sync_methods


@add_sync_methods
class Client(ABC):
    # pylint: disable=too-many-public-methods
    @abstractmethod
    async def get_block(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> Union[StarknetBlock, PreConfirmedStarknetBlock]:
        """
        Retrieve the block's data by its number or hash

        Alias of :meth:`get_block_with_txs`.

        :param block_hash: Block's hash or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :param block_number: Block's number or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :return: StarknetBlock object representing retrieved block
        """

    @abstractmethod
    async def get_block_with_txs(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> Union[StarknetBlock, PreConfirmedStarknetBlock]:
        """
        Retrieve the block's data by its number or hash.

        :param block_hash: Block's hash or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :param block_number: Block's number or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :return: StarknetBlock object representing retrieved block with transactions.
        """

    @abstractmethod
    async def get_block_with_tx_hashes(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> Union[StarknetBlockWithTxHashes, PreConfirmedStarknetBlockWithTxHashes]:
        """
        Retrieve the block's data with a list of contained transaction hashes.

        :param block_hash: Block's hash or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :param block_number: Block's number or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :return: StarknetBlockWithTxHashes object representing retrieved block with transactions.
        """

    @abstractmethod
    async def get_block_with_receipts(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> Union[StarknetBlockWithReceipts, PreConfirmedStarknetBlockWithReceipts]:
        """
        Retrieve the block's data with a list of receipts for contained transactions.

        :param block_hash: Block's hash or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :param block_number: Block's number or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :return: StarknetBlockWithReceipts object representing retrieved block with transactions.
        """

    @abstractmethod
    async def trace_block_transactions(
        self,
        block_hash: Optional[Union[Hash, LatestTag]] = None,
        block_number: Optional[Union[int, LatestTag]] = None,
    ) -> List[BlockTransactionTrace]:
        """
        Receive the traces of all the transactions within specified block

        :param block_hash: Block's hash
        :param block_number: Block's number or "pre_confirmed" for pre_confirmed block
        :return: BlockTransactionTraces object representing received traces
        """

    @abstractmethod
    async def get_state_update(
        self,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> Union[BlockStateUpdate, PreConfirmedBlockStateUpdate]:
        """
        Get the information about the result of executing the requested block

        :param block_hash: Block's hash or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :param block_number: Block's number or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :return: BlockStateUpdate object representing changes in the requested block
        """

    @abstractmethod
    async def get_storage_at(
        self,
        contract_address: Hash,
        key: int,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> int:
        """
        :param contract_address: Contract's address on Starknet
        :param key: An address of the storage variable inside the contract.
        :param block_hash: Block's hash or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :param block_number: Block's number or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :return: Storage value of given contract
        """

    @abstractmethod
    async def get_storage_proof(
        self,
        block_hash: Optional[Union[Hash, LatestTag]] = None,
        block_number: Optional[Union[int, LatestTag]] = None,
        class_hashes: Optional[List[int]] = None,
        contract_addresses: Optional[List[int]] = None,
        contracts_storage_keys: Optional[List[ContractsStorageKeys]] = None,
    ) -> StorageProofResponse:
        """
        Get merkle paths in one of the state tries: global state, classes, individual contract.

        :param block_hash: Block's hash or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :param block_number: Block's number or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :param class_hashes: List of the class hashes for which we want to prove membership in the classes trie.
        :param contract_addresses: List of the contract addresses for which we want to prove membership in the
                                    contracts trie.
        :param contracts_storage_keys: List of the contract address and storage keys pairs.
        :return: StorageProofResponse object.
        """

    @abstractmethod
    async def get_transaction(
        self,
        tx_hash: Hash,
    ) -> Transaction:
        """
        Get the details and status of a submitted transaction

        :param tx_hash: Transaction's hash
        :return: Transaction object
        """

    @abstractmethod
    async def get_transaction_receipt(
        self,
        tx_hash: Hash,
    ) -> TransactionReceiptWithBlockInfo:
        """
        Get the transaction receipt

        :param tx_hash: Transaction's hash
        :return: Transaction receipt object on Starknet
        """

    @abstractmethod
    async def get_transaction_status(self, tx_hash: Hash) -> TransactionStatusResponse:
        """
        Gets the transaction status (possibly reflecting that the transaction is still in the mempool,
        or dropped from it).

        :param tx_hash: Hash of the executed transaction.
        :return: Finality and execution status of a transaction.
        """

    # https://community.starknet.io/t/efficient-utilization-of-sequencer-capacity-in-starknet-v0-12-1/95607
    async def wait_for_tx(
        self,
        tx_hash: Hash,
        check_interval: float = 2,
        retries: int = 500,
    ) -> TransactionReceiptWithBlockInfo:
        # pylint: disable=too-many-branches
        """
        Awaits the transaction until its status is either ``ACCEPTED_ON_L2`` or ``ACCEPTED_ON_L1``
        and its receipt can be fetched.

        :param tx_hash: Transaction's hash.
        :param check_interval: Defines the interval between checks.
        :param retries: Defines how many times the transaction is checked until an error is thrown.
        :raises TransactionRevertedError: If the transaction execution status is ``REVERTED``.
        :raises TransactionNotReceivedError: If the transaction is not received within the given number of checks,
            i.e., its status never reaches ``ACCEPTED_ON_L2`` or ``ACCEPTED_ON_L1``.
        :return: Transaction receipt.
        """
        if check_interval <= 0:
            raise ValueError("Argument check_interval has to be greater than 0.")
        if retries <= 0:
            raise ValueError("Argument retries has to be greater than 0.")

        transaction_received = False
        while True:
            retries -= 1
            try:
                if not transaction_received:
                    tx_status = await self.get_transaction_status(tx_hash=tx_hash)

                    if (
                        tx_status.execution_status
                        == TransactionExecutionStatus.REVERTED
                    ):
                        raise TransactionRevertedError(
                            message=(
                                tx_status.failure_reason
                                if tx_status.failure_reason is not None
                                else "Transaction reverted with unknown reason."
                            )
                        )

                    if tx_status.finality_status in (
                        TransactionStatus.ACCEPTED_ON_L2,
                        TransactionStatus.ACCEPTED_ON_L1,
                    ):
                        transaction_received = True
                else:
                    tx_receipt = await self.get_transaction_receipt(tx_hash=tx_hash)

                    if (
                        tx_receipt.execution_status
                        == TransactionExecutionStatus.REVERTED
                    ):
                        raise TransactionRevertedError(message=tx_receipt.revert_reason)
                    return tx_receipt

                if retries == 0:
                    raise TransactionNotReceivedError()

                await asyncio.sleep(check_interval)

            except asyncio.CancelledError as exc:
                raise TransactionNotReceivedError from exc
            except ClientError as exc:
                if exc.code != 29:
                    raise exc

                if retries == 0:
                    raise TransactionNotReceivedError from exc

                await asyncio.sleep(check_interval)

    @abstractmethod
    async def estimate_fee(
        self,
        tx: Union[AccountTransaction, List[AccountTransaction]],
        skip_validate: bool = False,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> Union[EstimatedFee, List[EstimatedFee]]:
        """
        Estimates the resources required by a given sequence of transactions when applied on a given state.
        If one of the transactions reverts or fails due to any reason (e.g. validation failure or an internal error),
        a TRANSACTION_EXECUTION_ERROR is returned.
        For v0-2 transactions the estimate is given in Wei, and for v3 transactions it is given in Fri.

        :param tx: Transaction to estimate
        :param skip_validate: Flag checking whether the validation part of the transaction should be executed.
        :param block_hash: Block's hash or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`.
        :param block_number: Block's number or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`.
        :return: Estimated amount of Wei executing specified transaction will cost.
        """

    @abstractmethod
    async def call_contract(
        self,
        call: Call,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> List[int]:
        """
        Call the contract with given instance of InvokeTransaction

        :param call: Call
        :param block_hash: Block's hash or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :param block_number: Block's number or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :return: List of integers representing contract's function output (structured like calldata)
        """

    @abstractmethod
    async def send_transaction(
        self,
        transaction: InvokeV3,
    ) -> SentTransactionResponse:
        """
        Send a transaction to the network

        :param transaction: Transaction object (i.e. Invoke).
        :return: SentTransactionResponse object
        """

    @abstractmethod
    async def deploy_account(
        self, transaction: DeployAccountV3
    ) -> DeployAccountTransactionResponse:
        """
        Deploy a pre-funded account contract to the network

        :param transaction: DeployAccount transaction
        :return: SentTransactionResponse object
        """

    @abstractmethod
    async def declare(self, transaction: DeclareV3) -> DeclareTransactionResponse:
        """
        Send a declare transaction

        :param transaction: Declare transaction
        :return: SentTransactionResponse object
        """

    @abstractmethod
    async def get_class_hash_at(
        self,
        contract_address: Hash,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> int:
        """
        Get the contract class hash for the contract deployed at the given address

        :param contract_address: Address of the contract whose class hash is to be returned
        :param block_hash: Block's hash or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :param block_number: Block's number or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :return: Class hash
        """

    @abstractmethod
    async def get_class_by_hash(
        self, class_hash: Hash
    ) -> Union[DeprecatedContractClass, SierraContractClass]:
        """
        Get the contract class for given class hash

        :param class_hash: Class hash
        :return: ContractClass object
        """

    @abstractmethod
    async def get_contract_nonce(
        self,
        contract_address: int,
        block_hash: Optional[Union[Hash, Tag]] = None,
        block_number: Optional[Union[int, Tag]] = None,
    ) -> int:
        """
        Get the latest nonce associated with the given address

        :param contract_address: Get the latest nonce associated with the given address
        :param block_hash: Block's hash or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :param block_number: Block's number or literals `"l1_accepted"`, `"pre_confirmed"` or `"latest"`
        :return: The last nonce used for the given contract
        """

    @abstractmethod
    async def get_chain_id(self) -> str:
        """Return the currently configured Starknet chain id"""

    @abstractmethod
    async def get_messages_status(self, transaction_hash: Hash) -> List[MessageStatus]:
        """
        Get L1 handler transaction data for all L1 to L2 messages sent by the given L1 transaction.

        :param transaction_hash: Hash of the L1 transaction
        :return: Status of the messages
        """

    @abstractmethod
    async def get_compiled_casm(self, class_hash: int) -> CasmClass:
        """
        Get the contract class definition in the given block associated with the given hash.

        :param class_hash: Hash of the contract class whose CASM will be returned
        :return: CasmClass object
        """

    @abstractmethod
    async def spec_version(self) -> str:
        """
        Returns the version of the Starknet JSON-RPC specification being used.

        :return: String with version of the Starknet JSON-RPC specification.
        """



from collections import OrderedDict

import pytest

from starknet_py.serialization.data_serializers.enum_serializer import EnumSerializer
from starknet_py.serialization.data_serializers.option_serializer import (
    OptionSerializer,
)
from starknet_py.serialization.data_serializers.struct_serializer import (
    StructSerializer,
)
from starknet_py.serialization.data_serializers.uint_serializer import UintSerializer

serializer = EnumSerializer(
    serializers=OrderedDict(
        a=UintSerializer(256),
        b=UintSerializer(128),
        c=StructSerializer(
            OrderedDict(
                my_option=OptionSerializer(UintSerializer(128)),
                my_uint=UintSerializer(256),
            )
        ),
    )
)


@pytest.mark.parametrize(
    "value, correct_serialized_value",
    [
        ({"a": 100}, [0, 100, 0]),
        ({"b": 200}, [1, 200]),
        ({"c": {"my_option": 300, "my_uint": 300}}, [2, 0, 300, 300, 0]),
    ],
)
def test_output_serializer(value, correct_serialized_value):
    deserialized = serializer.deserialize(correct_serialized_value)

    deserialized_and_serialized = serializer.serialize(deserialized)
    serialized_value = serializer.serialize(value)

    assert deserialized_and_serialized == correct_serialized_value
    assert serialized_value == correct_serialized_value
    assert serialized_value == deserialized_and_serialized


def test_serializer_throws_on_wrong_parameters():
    with pytest.raises(ValueError, match="Can serialize only one enum variant, got: 2"):
        serializer.serialize({"a": 100, "b": 200})

    with pytest.raises(ValueError, match="Can serialize only one enum variant, got: 0"):
        serializer.serialize({})



import pytest

from starknet_py.abi.v0 import AbiParser
from starknet_py.constants import DEFAULT_DEPLOYER_ADDRESS
from starknet_py.net.udc_deployer.deployer import _deployer_abi


@pytest.fixture(scope="package", autouse=True)
async def check_if_udc_is_deployed(client):
    class_hash = await client.get_class_hash_at(
        contract_address=DEFAULT_DEPLOYER_ADDRESS
    )

    assert isinstance(class_hash, int)
    assert class_hash != 0


@pytest.fixture(scope="package", autouse=True)
async def check_if_udc_has_expected_abi(gateway_client):
    code = await gateway_client.get_code(contract_address=DEFAULT_DEPLOYER_ADDRESS)

    assert AbiParser(code.abi).parse() == _deployer_abi
